from __future__ import annotations
import asyncio
from typing import AsyncIterator, Iterable, List, Optional, Sequence
import torch
from ..node.actors import HonestNodeActor, ByzantineNodeActor
from ...aggregators import Aggregator
from ...pre_aggregators import PreAggregator
import numpy as np
from ...engine.storage.shared_store import SharedTensorHandle, register_tensor, cleanup_tensor
from ...engine.graph.ops import make_single_operator_graph
from ...engine.graph.scheduler import NodeScheduler
from ...engine.graph.pool import ActorPool

class ParameterServer:
    """
    Parameter server for synchronous distributed training.

    This class orchestrates a parameter server training loop where:
    1. All honest nodes compute gradients in parallel
    2. Byzantine nodes generate malicious gradients (if present)
    3. Gradients are optionally pre-aggregated
    4. An aggregator combines all gradients into a single update
    5. The aggregated gradient is sent back to all nodes

    Parameters
    ----------
    honest_nodes : List[HonestNodeActor]
        List of honest node actors that compute legitimate gradients.
    byzantine_nodes : List[ByzantineNodeActor]
        List of Byzantine node actors that generate malicious gradients.
    aggregator : Aggregator
        Aggregator instance to combine gradients (e.g., MultiKrum, Median).
    pre_aggregator : Optional[PreAggregator], optional
        Optional pre-aggregator to transform gradients before aggregation
        (e.g., Bucketing, Clipping).
    update_byzantines : bool, optional
        If True, Byzantine nodes also receive the aggregated gradient update.
        Default is False.
    actor_pool : ActorPool | None, optional
        Optional actor pool for parallel aggregation. If None, aggregation
        runs on the main thread.
    scheduler_metadata : Optional[dict], optional
        Additional metadata to pass to the graph scheduler.

    Examples
    --------
    >>> from byzpy.aggregators.geometric_wise.krum import MultiKrum
    >>> aggregator = MultiKrum(f=2, q=5)
    >>> ps = ParameterServer(
    ...     honest_nodes=honest_actors,
    ...     byzantine_nodes=byz_actors,
    ...     aggregator=aggregator
    ... )
    >>> for _ in range(100):
    ...     await ps.round()
    >>> await ps.shutdown()
    """
    def __init__(
        self,
        honest_nodes: List[HonestNodeActor],
        byzantine_nodes: List[ByzantineNodeActor],
        aggregator: Aggregator,
        pre_aggregator: Optional[PreAggregator] = None,
        update_byzantines: bool = False,
        *,
        actor_pool: ActorPool | None = None,
        scheduler_metadata: Optional[dict] = None,
    ):
        self.hon = honest_nodes
        self.byz = byzantine_nodes
        self.agg = aggregator
        self.pre = pre_aggregator
        self.update_byz = update_byzantines
        self.pool = actor_pool
        self.scheduler: NodeScheduler | None = None
        if actor_pool is not None:
            graph = make_single_operator_graph(
                node_name="agg",
                operator=self.agg,
                input_keys=("gradients",),
            )
            self.scheduler = NodeScheduler(graph, pool=actor_pool, metadata=scheduler_metadata)

    async def _stream_honest(self) -> AsyncIterator[torch.Tensor]:
        coros = [h.honest_gradient_for_next_batch() for h in self.hon]
        for fut in asyncio.as_completed(coros):
            yield await fut

    async def _stream_byz(self, honest_grads: Sequence[torch.Tensor]) -> AsyncIterator[torch.Tensor]:
        if not self.byz:
            return
        coros = [b.byzantine_gradient_for_next_batch(honest_grads) for b in self.byz]
        for fut in asyncio.as_completed(coros):
            yield await fut

    async def round(self) -> torch.Tensor:
        """
        Execute one training round.

        This method:
        1. Collects gradients from all honest nodes
        2. Generates malicious gradients from Byzantine nodes
        3. Optionally applies pre-aggregation
        4. Aggregates all gradients
        5. Sends the aggregated gradient to all nodes

        Returns
        -------
        torch.Tensor
            The aggregated gradient vector.
        """
        gradients: list[torch.Tensor] = []

        async for grad in self._stream_honest():
            gradients.append(grad)

        async for grad in self._stream_byz(tuple(gradients)):
            gradients.append(grad)

        if self.pre:
            gradients = list(self.pre.pre_aggregate(gradients))

        handles = self._register_shared_gradients(gradients)
        try:
            if self.scheduler is not None:
                result = await self.scheduler.run({"gradients": handles})
                g = result["agg"]
            else:
                g = self.agg.aggregate(handles)
        finally:
            self._cleanup_shared_gradients(handles)

        await asyncio.gather(*[
            n.apply_server_gradient(g) for n in self.hon
        ] + ([
            n.apply_server_gradient(g) for n in self.byz
        ] if self.update_byz else []))
        return g

    async def shutdown(self):
        await asyncio.gather(*[n._ref._backend.close() for n in (self.hon + self.byz)])

    def _register_shared_gradients(self, grads: Sequence[torch.Tensor]) -> list[SharedTensorHandle]:
        handles: list[SharedTensorHandle] = []
        for grad in grads:
            arr = grad.detach().cpu().numpy()
            handles.append(register_tensor(np.array(arr, copy=True)))
        return handles

    def _cleanup_shared_gradients(self, handles: Sequence[SharedTensorHandle]) -> None:
        for handle in handles:
            cleanup_tensor(handle)



from __future__ import annotations

import time

from byzpy.engine.transport.tcp_simple import TcpMailbox, send_message


def test_tcp_mailbox_send_recv():
    mbox = TcpMailbox()
    try:
        addr = ("127.0.0.1", mbox.port)
        send_message(addr, {"hello": "world"})
        msg = mbox.recv(timeout=1.0)
        assert msg == {"hello": "world"}
    finally:
        mbox.close()


def test_tcp_mailbox_multiple_messages():
    mbox = TcpMailbox()
    try:
        addr = ("127.0.0.1", mbox.port)
        for i in range(3):
            send_message(addr, i)
        seen = []
        start = time.time()
        while len(seen) < 3 and (time.time() - start) < 2.0:
            seen.append(mbox.recv(timeout=1.0))
        assert sorted(seen) == [0, 1, 2]
    finally:
        mbox.close()



from __future__ import annotations

from typing import Any, Sequence

import numpy as np

from ...engine.storage.shared_store import SharedTensorHandle, open_tensor

try:  # optional torch dependency
    import torch
    _HAS_TORCH = True
except Exception:  # pragma: no cover
    torch = None  # type: ignore
    _HAS_TORCH = False


def flatten_gradients(gradients: Sequence[Any]) -> tuple[tuple[int, ...], np.ndarray]:
    arrays = [_to_numpy(g) for g in gradients]
    stacked = np.stack(arrays, axis=0)
    shape = stacked.shape[1:]
    flat = stacked.reshape(stacked.shape[0], -1)
    return shape, flat


def _to_numpy(grad: Any) -> np.ndarray:
    if _HAS_TORCH and isinstance(grad, torch.Tensor):  # type: ignore[arg-type]
        return grad.detach().cpu().numpy()
    if isinstance(grad, np.ndarray):
        return grad
    if isinstance(grad, SharedTensorHandle):
        with open_tensor(grad) as arr:
            return np.array(arr, copy=True)
    if isinstance(grad, dict) and {"name", "shape", "dtype"} <= grad.keys():
        handle = SharedTensorHandle(**grad)
        with open_tensor(handle) as arr:
            return np.array(arr, copy=True)
    return np.asarray(grad)


__all__ = ["flatten_gradients"]



from __future__ import annotations

import asyncio
import threading
from typing import Sequence

import pytest
import torch

from byzpy.aggregators.base import Aggregator
from byzpy.engine.graph.graph import ComputationGraph
from byzpy.engine.graph.ops import (
    RemoteCallableOp,
    make_single_operator_graph,
)
from byzpy.engine.graph.pool import ActorPool, ActorPoolConfig
from byzpy.engine.graph.scheduler import NodeScheduler


class _SumAggregator(Aggregator):
    def aggregate(self, gradients: Sequence[torch.Tensor]) -> torch.Tensor:
        total = torch.zeros_like(gradients[0])
        for grad in gradients:
            total = total + grad
        return total


@pytest.mark.asyncio
async def test_aggregator_op_single_node_graph():
    op = _SumAggregator()
    graph = make_single_operator_graph(
        node_name="agg",
        operator=op,
        input_keys=("gradients",),
    )

    pool = ActorPool([ActorPoolConfig(backend="thread", count=2)])
    scheduler = NodeScheduler(graph, pool=pool)

    grads = [torch.arange(4, dtype=torch.float32), torch.arange(4, dtype=torch.float32)]
    out = await scheduler.run({"gradients": grads})
    assert torch.equal(out["agg"], grads[0] + grads[1])

    await pool.shutdown()


@pytest.mark.asyncio
async def test_remote_callable_op_runs_on_pool_thread():
    thread_ids: list[int] = []

    def _fn(value: torch.Tensor) -> torch.Tensor:
        thread_ids.append(threading.get_ident())
        return value * 2

    op = RemoteCallableOp(_fn, input_mapping={"value": "value"})
    graph: ComputationGraph = make_single_operator_graph(
        node_name="remote",
        operator=op,
        input_keys=("value",),
    )

    pool = ActorPool([ActorPoolConfig(backend="thread", count=2)])
    scheduler = NodeScheduler(graph, pool=pool)

    main_thread = threading.get_ident()
    value = torch.ones(3, dtype=torch.float32)
    out = await scheduler.run({"value": value})

    assert torch.equal(out["remote"], value * 2)
    assert thread_ids, "remote callable should execute on pool worker"
    assert thread_ids[0] != main_thread

    await pool.shutdown()



"""
Category 6: Fully Asynchronous P2P Training Tests

Tests for fully decentralized P2P training where nodes progress independently
without a synchronous round() API.
"""
from __future__ import annotations

import asyncio
import pytest
import torch

from byzpy.engine.node.decentralized import DecentralizedNode
from byzpy.engine.node.cluster import DecentralizedCluster
from byzpy.engine.node.application import HonestNodeApplication
from byzpy.engine.node.context import InProcessContext, ProcessContext
from byzpy.engine.peer_to_peer.topology import Topology
from byzpy.engine.graph.pool import ActorPoolConfig
from byzpy.engine.graph.ops import CallableOp, make_single_operator_graph
from byzpy.engine.graph.graph import GraphInput
from byzpy.engine.graph.graph import ComputationGraph, GraphNode
from byzpy.aggregators.coordinate_wise import CoordinateWiseMedian


@pytest.mark.asyncio
async def test_nodes_progress_independently_without_round_api():
    """Verify nodes can progress independently without central round() coordination."""
    topology = Topology.ring(3, k=1)
    cluster = DecentralizedCluster()
    received_gradients = {f"node{i}": [] for i in range(3)}

    # Create nodes with message-driven pipelines
    for i in range(3):
        app = HonestNodeApplication(
            name=f"node{i}",
            actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        )

        # Half-step pipeline (triggered by local state/timer)
        def make_half_step(node_idx):
            def half_step(lr):
                return torch.tensor([float(node_idx) * lr, 2.0 * float(node_idx) * lr])
            return half_step

        half_step_graph = make_single_operator_graph(
            node_name="half_step",
            operator=CallableOp(make_half_step(i), input_mapping={"lr": "lr"}),
            input_keys=("lr",),
        )
        app.register_pipeline("half_step", half_step_graph)

        # Aggregation pipeline
        aggregator = CoordinateWiseMedian()
        aggregate_graph = make_single_operator_graph(
            node_name="aggregate",
            operator=aggregator,
            input_keys=("gradients",),
        )
        app.register_pipeline("aggregate", aggregate_graph)

        node = await cluster.add_node(
            node_id=f"node{i}",
            application=app,
            topology=topology,
            context=InProcessContext(),
        )

        # Register message handler that collects gradients
        def make_aggregation_handler(nid):
            async def on_gradient(from_id, payload):
                received_gradients[nid].append(payload["vector"])
            return on_gradient

        node.register_message_handler("gradient", make_aggregation_handler(f"node{i}"))

    cluster._update_node_id_maps()
    await cluster.start_all()
    await asyncio.sleep(0.3)

    # Each node independently triggers half-step
    for node_id, node in cluster.nodes.items():
        half_result = await node.execute_pipeline("half_step", {"lr": 0.01})
        # Broadcast to neighbors
        await node.broadcast_message("gradient", {"vector": half_result["half_step"]})

    # Wait for messages to propagate
    await asyncio.sleep(0.5)

    # Verify each node received gradients from neighbors
    for node_id in cluster.nodes:
        assert len(received_gradients[node_id]) >= 1  # At least one neighbor

    await cluster.shutdown_all()


@pytest.mark.asyncio
async def test_nodes_trigger_half_step_independently():
    """Verify nodes can trigger half-step pipelines independently based on local state."""
    app = HonestNodeApplication(
        name="test",
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
    )

    def half_step(lr):
        return torch.tensor([1.0 * lr, 2.0 * lr])

    graph = make_single_operator_graph(
        node_name="half_step",
        operator=CallableOp(half_step, input_mapping={"lr": "lr"}),
        input_keys=("lr",),
    )
    app.register_pipeline("half_step", graph)

    node = DecentralizedNode(
        node_id="node0",
        application=app,
        context=InProcessContext(),
    )

    await node.start()

    # Node independently triggers half-step (e.g., based on timer or local state)
    result1 = await node.execute_pipeline("half_step", {"lr": 0.01})
    await asyncio.sleep(0.1)
    result2 = await node.execute_pipeline("half_step", {"lr": 0.01})

    assert "half_step" in result1
    assert "half_step" in result2

    await node.shutdown()


@pytest.mark.asyncio
async def test_nodes_aggregate_when_neighbors_respond():
    """Verify nodes automatically trigger aggregation when they receive neighbor messages."""
    topology = Topology.ring(3, k=1)
    cluster = DecentralizedCluster()
    aggregation_triggered = {"node0": False}

    app = HonestNodeApplication(
        name="node0",
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
    )

    # Aggregation pipeline
    aggregator = CoordinateWiseMedian()
    aggregate_graph = make_single_operator_graph(
        node_name="aggregate",
        operator=aggregator,
        input_keys=("gradients",),
    )
    app.register_pipeline("aggregate", aggregate_graph)

    node = await cluster.add_node(
        node_id="node0",
        application=app,
        topology=topology,
        context=InProcessContext(),
    )

    # Message handler triggers aggregation when messages arrive
    async def on_gradient(from_id, payload):
        # Trigger aggregation pipeline
        gradients = [payload["vector"]]
        result = await node.execute_pipeline("aggregate", {"gradients": gradients})
        aggregation_triggered["node0"] = True

    node.register_message_handler("gradient", on_gradient)

    cluster._update_node_id_maps()
    await cluster.start_all()
    await asyncio.sleep(0.2)

    # Simulate receiving gradients (in real scenario, these come from neighbors)
    # For this test, we'll directly test aggregation
    gradients = [
        torch.tensor([1.0, 2.0]),
        torch.tensor([2.0, 3.0]),
    ]
    result = await node.execute_pipeline("aggregate", {"gradients": gradients})
    aggregation_triggered["node0"] = True

    assert aggregation_triggered["node0"]
    assert "aggregate" in result

    await cluster.shutdown_all()



@pytest.mark.asyncio
async def test_p2p_training_fully_asynchronous_ring():
    """End-to-end fully asynchronous P2P training with ring topology."""
    topology = Topology.ring(4, k=1)
    cluster = DecentralizedCluster()
    training_rounds = {f"node{i}": 0 for i in range(4)}

    # Create nodes with fully asynchronous training logic
    for i in range(4):
        app = HonestNodeApplication(
            name=f"node{i}",
            actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        )

        # Half-step pipeline
        def make_half_step(node_idx):
            def half_step(lr):
                return torch.tensor([float(node_idx) * lr, 2.0 * float(node_idx) * lr])
            return half_step

        half_step_graph = make_single_operator_graph(
            node_name="half_step",
            operator=CallableOp(make_half_step(i), input_mapping={"lr": "lr"}),
            input_keys=("lr",),
        )
        app.register_pipeline("half_step", half_step_graph)

        # Aggregation pipeline
        aggregator = CoordinateWiseMedian()
        aggregate_graph = make_single_operator_graph(
            node_name="aggregate",
            operator=aggregator,
            input_keys=("gradients",),
        )
        app.register_pipeline("aggregate", aggregate_graph)

        node = await cluster.add_node(
            node_id=f"node{i}",
            application=app,
            topology=topology,
            context=InProcessContext(),
        )

        # Message handler: when gradient received, trigger aggregation
        def make_training_handler(nid, n):
            async def on_gradient(from_id, payload):
                # Trigger aggregation
                training_rounds[nid] += 1
            return on_gradient

        node.register_message_handler("gradient", make_training_handler(f"node{i}", node))

    cluster._update_node_id_maps()
    await cluster.start_all()
    await asyncio.sleep(0.5)

    # Each node independently starts training
    for node_id, node in cluster.nodes.items():
        # Trigger half-step
        result = await node.execute_pipeline("half_step", {"lr": 0.01})
        # Broadcast to neighbors
        await node.broadcast_message("gradient", {"vector": result["half_step"]})

    # Wait for asynchronous training to progress
    await asyncio.sleep(1.0)

    # Verify nodes made progress independently
    for node_id in training_rounds:
        assert training_rounds[node_id] > 0

    await cluster.shutdown_all()




from __future__ import annotations
import asyncio
import traceback
import uuid
from typing import Any, Dict, Optional

import cloudpickle

from ..base import ActorBackend
from ..channels import Endpoint
from ..router import channel_router
from ..transports import ucx, tcp


def _resolve_backend(scheme: str, actor_id: str):
    return channel_router.resolve(scheme, actor_id)

# ----------------------------- GPU local backend --------------------------------

class GPUActorBackend(ActorBackend):
    """
    Local, single-process actor that prefers CUDA tensors, exposes endpoint scheme 'gpu'.
    Mailboxes are in-process asyncio. Cross-backend routing:
      - to 'gpu' in same process: enqueue (zero-copy PyTorch tensors)
      - to 'thread'/'process': delegate to their registries
      - to 'tcp': use your TCP server protocol
      - to 'ucx': UCX fast path (GPU→GPU) using UCX-Py
    """
    def __init__(self) -> None:
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._obj = None
        self._actor_id = str(uuid.uuid4())
        self._queues: Dict[str, asyncio.Queue] = {}
        channel_router.register("gpu", self._actor_id, self)

    async def start(self):
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

    async def construct(self, cls_or_factory: Any, *, args: tuple, kwargs: dict) -> None:
        target = cls_or_factory
        self._obj = target(*args, **kwargs) if not isinstance(target, type) else target(*args, **kwargs)

    async def call(self, method: str, *args, **kwargs):
        fn = getattr(self._obj, method)
        if asyncio.iscoroutinefunction(fn):
            return await fn(*args, **kwargs)
        loop = self._loop or asyncio.get_running_loop()
        return await loop.run_in_executor(None, lambda: fn(*args, **kwargs))

    async def close(self) -> None:
        channel_router.unregister("gpu", self._actor_id)

    # ---- decentralized channels ----
    async def get_endpoint(self) -> Endpoint:
        return Endpoint(scheme="gpu", address="", actor_id=self._actor_id)

    async def chan_open(self, name: str) -> Endpoint:
        if name not in self._queues:
            self._queues[name] = asyncio.Queue()
        return await self.get_endpoint()

    async def _deliver_local(self, name: str, from_ep: Endpoint, payload: Any) -> None:
        q = self._queues.setdefault(name, asyncio.Queue())
        await q.put((from_ep, payload))

    async def chan_put(self, *, from_ep: Endpoint, to_ep: Endpoint, name: str, payload: Any) -> None:
        # 1) target == local gpu actor
        if to_ep.scheme == "gpu":
            if to_ep.actor_id == self._actor_id:
                await self._deliver_local(name, from_ep, payload)
                return
            peer = _resolve_backend("gpu", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local gpu actor {to_ep.actor_id}")
            await peer._deliver_local(name, from_ep, payload)
            return

        # 2) route to local thread/process
        if to_ep.scheme == "thread":
            peer = _resolve_backend("thread", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return

        if to_ep.scheme == "process":
            peer = _resolve_backend("process", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return

        # 3) UCX fast path (GPU→GPU) to remote UCX server
        if to_ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port = to_ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                })
                tag, desc = ucx.pack_payload(payload)
                await ucx.send_payload(e, tag, desc, payload)
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)

            await ucx.call(host, int(port), _op)
            return

        # 4) TCP fallback
        if to_ep.scheme == "tcp":
            await tcp.chan_put(
                to_ep.address,
                from_ep=from_ep,
                to_ep=to_ep,
                name=name,
                payload=payload,
            )
            return

        raise RuntimeError(f"GPUActorBackend cannot route to {to_ep.scheme!r}")

    async def chan_get(self, *, ep: Endpoint, name: str, timeout: Optional[float]):
        # 1) local GPU mailbox
        if ep.scheme == "gpu" and ep.actor_id == self._actor_id:
            q = self._queues.setdefault(name, asyncio.Queue())
            if timeout is None:
                _, payload = await q.get()
                return payload
            try:
                _, payload = await asyncio.wait_for(q.get(), timeout=timeout)
                return payload
            except asyncio.TimeoutError:
                return None

        # 2) read from local thread/process backends (same Python runtime)
        if ep.scheme == "thread":
            peer = _resolve_backend("thread", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)

        if ep.scheme == "process":
            peer = _resolve_backend("process", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)

        # 3) read from a UCX server mailbox
        if ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port = ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {"op": "chan_get", "name": name, "timeout": timeout, "actor_id": ep.actor_id})
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
                if rep.get("payload_is_none", False):
                    return None
                return await ucx.recv_payload(e)

            return await ucx.call(host, int(port), _op)

        # 4) TCP read
        if ep.scheme == "tcp":
            return await tcp.chan_get(
                ep.address,
                name=name,
                timeout=timeout,
                actor_id=ep.actor_id,
            )

        raise RuntimeError("Endpoint mismatch")

# ----------------------------- UCX remote (optional) ----------------------------

class UCXRemoteActorBackend(ActorBackend):
    """
    Client for a UCXRemoteActorServer. Lets you do:
      backend = UCXRemoteActorBackend(host, port)
      ref = ActorRef(backend)  # normal construct/call
    Also provides UCX-aware chan_put/chan_get for GPU GPU transfers.
    """
    def __init__(self, host: str, port: int):
        self._host, self._port = host, int(port)
        self._ep = None
        self._actor_id: Optional[str] = None
        self._io_lock = asyncio.Lock()
        self._advertised_address: Optional[str] = None

    async def start(self) -> None:
        if self._ep is None:
            if not ucx.have_ucx():
                raise RuntimeError("UCX is not available (install ucxx or ucp)")
            # Persistent endpoint dedicated to this backend (no sharing).
            self._ep = await ucx.create_endpoint(self._host, self._port)

    async def close(self) -> None:
        async with self._io_lock:
            if self._ep is None:
                self._actor_id = None
                return
            try:
                if self._actor_id is not None:
                    await ucx.send_control(self._ep, {"op": "close", "actor_id": self._actor_id})
                    await ucx.recv_control(self._ep)
            except Exception:
                pass
            finally:
                try:
                    await self._ep.close()
                except Exception:
                    pass
                self._ep = None
                self._actor_id = None
                self._advertised_address = None

    async def construct(self, cls_or_factory: Any, *, args: tuple, kwargs: dict) -> None:
        async with self._io_lock:
            await self.start()
            blob = cloudpickle.dumps(cls_or_factory)
            await ucx.send_control(self._ep, {"op": "construct", "blob": blob, "args": args, "kwargs": kwargs})
            rep = await ucx.recv_control(self._ep)
            if not rep.get("ok", False):
                et, em, tb = rep["payload"]
                raise RuntimeError(f"[UCXRemote construct {et}] {em}\n{tb}")
            self._actor_id = rep["actor_id"]

    async def call(self, method: str, *args, **kwargs):
        async with self._io_lock:
            if not self._actor_id:
                raise RuntimeError("UCX remote actor not constructed.")
            await self.start()
            await ucx.send_control(self._ep, {
                "op": "call",
                "actor_id": self._actor_id,
                "method": method,
                "args": args,
                "kwargs": kwargs,
            })
            rep = await ucx.recv_control(self._ep)
            if rep.get("ok", False):
                return rep["payload"]
            et, em, tb = rep["payload"]
            raise RuntimeError(f"[UCXRemote call {et}] {em}\n{tb}")

    async def get_endpoint(self) -> Endpoint:
        async with self._io_lock:
            await self.start()
            await ucx.send_control(self._ep, {"op": "get_ep"})
            rep = await ucx.recv_control(self._ep)
            if not rep.get("ok", False):
                raise RuntimeError(rep)
            payload = rep["payload"]
            addr = payload.get("address")
            if addr:
                self._advertised_address = addr
            return Endpoint(**payload)

    async def chan_open(self, name: str) -> Endpoint:
        async with self._io_lock:
            await self.start()
            await ucx.send_control(self._ep, {"op": "chan_open", "name": name})
            rep = await ucx.recv_control(self._ep)
            if not rep.get("ok", False):
                raise RuntimeError(rep)
            payload = rep["payload"]
            addr = payload.get("address")
            if addr:
                self._advertised_address = addr
            return Endpoint(**payload)

    def _is_same_server(self, address: str) -> bool:
        if self._advertised_address and address == self._advertised_address:
            return True
        try:
            host, port_str = address.rsplit(":", 1)
            port = int(port_str)
        except ValueError:
            return False
        if port != self._port:
            return False
        if host == self._host:
            return True
        if self._host in ("0.0.0.0", "") and host in ("127.0.0.1", "localhost"):
            return True
        if host in ("0.0.0.0", "") and self._host in ("127.0.0.1", "localhost"):
            return True
        return False

    async def chan_put(self, *, from_ep: Endpoint, to_ep: Endpoint, name: str, payload: Any) -> None:
        # route to local thread/process/gpu
        if to_ep.scheme == "thread":
            peer = _resolve_backend("thread", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return
        if to_ep.scheme == "process":
            peer = _resolve_backend("process", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return
        if to_ep.scheme == "gpu":
            peer = _resolve_backend("gpu", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local gpu actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return

        # same UCX server via persistent ep (already serialized by _io_lock)
        if to_ep.scheme == "ucx" and self._is_same_server(to_ep.address):
            async with self._io_lock:
                await self.start()
                await ucx.send_control(self._ep, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                })
                tag, desc = ucx.pack_payload(payload)
                await ucx.send_payload(self._ep, tag, desc, payload)
                rep = await ucx.recv_control(self._ep)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
            return

        # different UCX server via pooled ep
        if to_ep.scheme == "ucx":
            host, port = to_ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                })
                tag, desc = ucx.pack_payload(payload)
                await ucx.send_payload(e, tag, desc, payload)
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)

            await ucx.call(host, int(port), _op)
            return

        # tcp fallback
        if to_ep.scheme == "tcp":
            await tcp.chan_put(
                to_ep.address,
                from_ep=from_ep,
                to_ep=to_ep,
                name=name,
                payload=payload,
            )
            return

        raise RuntimeError(f"UCXRemoteActorBackend cannot route to {to_ep.scheme!r}")

    async def chan_get(self, *, ep: Endpoint, name: str, timeout: Optional[float]):
        # local thread/process/gpu reads
        if ep.scheme == "thread":
            peer = _resolve_backend("thread", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)
        if ep.scheme == "process":
            peer = _resolve_backend("process", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)
        if ep.scheme == "gpu":
            peer = _resolve_backend("gpu", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local gpu actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)

        # same UCX server via persistent ep
        if ep.scheme == "ucx" and self._is_same_server(ep.address):
            async with self._io_lock:
                await self.start()
                await ucx.send_control(self._ep, {"op": "chan_get", "name": name, "timeout": timeout})
                rep = await ucx.recv_control(self._ep)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
                if rep.get("payload_is_none", False):
                    return None
                return await ucx.recv_payload(self._ep)

        # different UCX server
        if ep.scheme == "ucx":
            host, port = ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {"op": "chan_get", "name": name, "timeout": timeout, "actor_id": ep.actor_id})
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
                if rep.get("payload_is_none", False):
                    return None
                return await ucx.recv_payload(e)

            return await ucx.call(host, int(port), _op)

        # tcp
        if ep.scheme == "tcp":
            return await tcp.chan_get(
                ep.address,
                name=name,
                timeout=timeout,
                actor_id=ep.actor_id,
            )

        raise RuntimeError("Endpoint mismatch")

# ----------------------------- UCX remote server --------------------------------

class UCXRemoteActorServer:
    """
    UCX multi-actor server with CUDA-aware channels.
    """
    def __init__(self, host: str = "0.0.0.0", port: int = 0):
        if not ucx.have_ucx():
            raise RuntimeError("UCX (ucp) is not available in this environment.")
        self._host, self._port = host, int(port)
        self._actors: Dict[str, Any] = {}
        self._mb: Dict[str, Dict[str, asyncio.Queue]] = {}
        self._listener = None
        self._addr_cache: Optional[str] = None

    def address(self) -> str:
        return self._addr_cache or f"{self._host}:{self._port}"

    async def _send_error(self, ep, exc: Exception) -> None:
        tb = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
        payload = (type(exc).__name__, str(exc), tb)
        try:
            await ucx.send_control(ep, {"ok": False, "payload": payload})
        except Exception:
            pass

    def _log_error(self, op: Any, exc: Exception) -> None:
        try:
            opname = op if isinstance(op, str) else repr(op)
            print(f"[UCXRemoteActorServer] error during {opname}: {type(exc).__name__}: {exc}")
        except Exception:
            pass

    async def serve(self):
        async def _handle(ep):
            bound_aid: Optional[str] = None
            try:
                while True:
                    req = await ucx.recv_control(ep)
                    op = req.get("op")

                    try:
                        if op == "construct":
                            target = cloudpickle.loads(req["blob"])
                            args = req.get("args", ())
                            kwargs = req.get("kwargs", {}) or {}
                            obj = target(*args, **kwargs)
                            aid = str(uuid.uuid4())
                            self._actors[aid] = obj
                            bound_aid = aid
                            await ucx.send_control(ep, {"ok": True, "actor_id": aid})

                        elif op == "get_ep":
                            if not bound_aid:
                                raise RuntimeError("no actor on this connection")
                            await ucx.send_control(ep, {"ok": True,
                                "payload": {"scheme": "ucx", "address": self.address(), "actor_id": bound_aid}})

                        elif op == "chan_open":
                            if not bound_aid:
                                raise RuntimeError("no actor on this connection")
                            name = req["name"]
                            await self._ensure_mb(bound_aid, name)
                            await ucx.send_control(ep, {"ok": True,
                                "payload": {"scheme": "ucx", "address": self.address(), "actor_id": bound_aid}})

                        elif op == "chan_put":
                            to = req["to"]; name = req["name"]
                            to_aid = to["actor_id"]
                            q = await self._ensure_mb(to_aid, name)
                            payload = await ucx.recv_payload(ep)
                            await q.put((req["from"], payload))
                            await ucx.send_control(ep, {"ok": True})

                        elif op == "chan_get":
                            aid = bound_aid or req.get("actor_id")
                            if not aid:
                                raise RuntimeError("actor_id missing for chan_get")
                            name = req["name"]; timeout = req.get("timeout")
                            q = await self._ensure_mb(aid, name)
                            try:
                                if timeout is None:
                                    fr, payload = await q.get()
                                else:
                                    fr, payload = await asyncio.wait_for(q.get(), timeout=timeout)
                                await ucx.send_control(ep, {"ok": True, "payload_is_none": False})
                                tag, desc = ucx.pack_payload(payload)
                                await ucx.send_payload(ep, tag, desc, payload)
                            except asyncio.TimeoutError:
                                await ucx.send_control(ep, {"ok": True, "payload_is_none": True})

                        elif op == "call":
                            if not bound_aid:
                                raise RuntimeError("no actor on this connection")
                            obj = self._actors[bound_aid]
                            import inspect
                            if inspect.iscoroutinefunction(getattr(obj, req["method"])):
                                res = await getattr(obj, req["method"])(*req.get("args", ()), **(req.get("kwargs", {}) or {}))
                            else:
                                loop = asyncio.get_running_loop()
                                res = await loop.run_in_executor(
                                    None,
                                    lambda: getattr(obj, req["method"])(*req.get("args", ()), **(req.get("kwargs", {}) or {}))
                                )
                            await ucx.send_control(ep, {"ok": True, "payload": res})

                        elif op == "close":
                            if bound_aid:
                                self._actors.pop(bound_aid, None)
                                self._mb.pop(bound_aid, None)
                            await ucx.send_control(ep, {"ok": True})

                        else:
                            raise RuntimeError(f"unknown op {op}")
                    except Exception as exc:
                        self._log_error(op, exc)
                        await self._send_error(ep, exc)
                        if op == "construct":
                            # construction failures leave the connection unbound
                            bound_aid = None

            except Exception:
                # best-effort error reply already sent above if possible
                pass
            finally:
                try:
                    await ep.close()
                except Exception:
                    pass

        def _host_ip() -> str:
            import socket
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s.connect(("8.8.8.8", 80))
                ip = s.getsockname()[0]; s.close()
                return ip
            except Exception:
                return "127.0.0.1"

        self._listener = ucx.create_listener(_handle, host=self._host, port=self._port)
        host = self._host if self._host not in ("0.0.0.0", "") else _host_ip()
        self._addr_cache = f"{host}:{int(self._listener.port)}"
        print(f"[UCXRemoteActorServer] Serving on ucx://{self._addr_cache}")
        while True:
            await asyncio.sleep(3600)

    async def _ensure_mb(self, aid: str, name: str) -> asyncio.Queue:
        chs = self._mb.setdefault(aid, {})
        q = chs.get(name)
        if q is None:
            q = asyncio.Queue()
            chs[name] = q
        return q

async def start_ucx_actor_server(host: str = "0.0.0.0", port: int = 0):
    srv = UCXRemoteActorServer(host=host, port=port)
    await srv.serve()



import torch

from byzpy.attacks.mimic import MimicAttack
from byzpy.engine.graph.operator import OpContext


def test_mimic_direct_expected():
    grads = [torch.tensor([1.0, 2.0]), torch.tensor([3.0, 4.0]), torch.tensor([5.0, 6.0])]
    attack = MimicAttack(epsilon=1)
    out = attack.apply(honest_grads=grads)
    assert torch.allclose(out, grads[1])


def test_mimic_chunk_matches_direct():
    grads = [torch.randn(128) for _ in range(5)]
    direct = MimicAttack(epsilon=3, chunk_size=32).apply(honest_grads=grads)

    chunked = MimicAttack(epsilon=3, chunk_size=32)
    inputs = {"honest_grads": grads}
    ctx = OpContext(node_name="mimic", metadata={"pool_size": 4})
    subtasks = list(chunked.create_subtasks(inputs, context=ctx))
    partials = [t.fn(*t.args, **t.kwargs) for t in subtasks]
    reduced = chunked.reduce_subtasks(partials, inputs, context=OpContext(node_name="mimic", metadata={}))

    assert torch.allclose(direct, reduced, atol=1e-7)



from __future__ import annotations
import asyncio
import contextlib
import math
import socket
import threading
from typing import Tuple

import pytest
import pytest_asyncio
import torch

from byzpy.engine.actor.base import ActorRef
from byzpy.engine.actor.channels import open_channel
from byzpy.engine.actor.backends.thread import ThreadActorBackend
from byzpy.engine.actor.backends.process import ProcessActorBackend
from byzpy.engine.actor.backends.remote import (
    RemoteActorBackend,
    start_actor_server,
)
from byzpy.engine.actor.backends.gpu import (
    GPUActorBackend,
    UCXRemoteActorBackend,
    start_ucx_actor_server,
)
from byzpy.engine.actor.transports.ucx import _ucx_mod


def _have_cuda() -> bool:
    try:
        return torch.cuda.is_available()
    except Exception:
        return False

def _have_cupy() -> bool:
    try:
        import cupy as _
        return True
    except Exception:
        return False

def _have_ucx() -> bool:
    return _ucx_mod() is not None


CUDA_OK = _have_cuda()
CUPY_OK = _have_cupy()
UCX_OK = _have_ucx()
UCX_GPU_OK = UCX_OK and CUDA_OK and CUPY_OK


class Worker:
    def __init__(self, bias: float = 0.0):
        self.bias = float(bias)

    def make(self, n: int) -> torch.Tensor:
        return torch.arange(n, dtype=torch.float32).contiguous()

    def score(self, x: torch.Tensor) -> float:
        return float((x * x).sum().item() + self.bias)


async def _wait_until_listening_tcp(host: str, port: int, timeout: float = 3.0) -> None:
    """Poll-connect until the TCP server accepts, or timeout."""
    deadline = asyncio.get_event_loop().time() + timeout
    last_err = None
    while True:
        try:
            r, w = await asyncio.open_connection(host, port)
            w.close()
            with contextlib.suppress(Exception):
                await w.wait_closed()
            return
        except Exception as e:
            last_err = e
            if asyncio.get_event_loop().time() >= deadline:
                raise TimeoutError(f"TCP server {host}:{port} did not start: {last_err!r}")
            await asyncio.sleep(0.05)


async def _wait_until_listening_ucx(host: str, port: int, timeout: float = 5.0) -> None:
    """Poll-create UCX endpoint until the UCX server accepts, or timeout."""
    m = _ucx_mod()
    if m is None:
        raise RuntimeError("UCX module not available")
    deadline = asyncio.get_event_loop().time() + timeout
    last_err = None
    while True:
        try:
            ep = await m.create_endpoint(host, int(port))
            with contextlib.suppress(Exception):
                await ep.close()
            return
        except Exception as e:
            last_err = e
            if asyncio.get_event_loop().time() >= deadline:
                raise TimeoutError(f"UCX server {host}:{port} did not start: {last_err!r}")
            await asyncio.sleep(0.05)


@pytest_asyncio.fixture(scope="module")
async def tcp_server_addr():
    """
    Run RemoteActorServer on a dedicated event loop in a background thread.
    """
    host = "127.0.0.1"
    with socket.socket() as s:
        s.bind((host, 0))
        port = s.getsockname()[1]

    started_evt = threading.Event()
    stop_evt = threading.Event()

    def _server_thread():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _runner():
            started_evt.set()
            await start_actor_server(host, port)

        task = loop.create_task(_runner())

        def _stop():
            if not task.done():
                task.cancel()

        def _watch_stop():
            stop_evt.wait()
            loop.call_soon_threadsafe(_stop)
            loop.call_soon_threadsafe(loop.stop)

        watcher = threading.Thread(target=_watch_stop, daemon=True)
        watcher.start()

        try:
            loop.run_forever()
        finally:
            with contextlib.suppress(asyncio.CancelledError):
                loop.run_until_complete(task)
            loop.close()

    th = threading.Thread(target=_server_thread, daemon=True)
    th.start()

    started_evt.wait(timeout=2.0)
    await _wait_until_listening_tcp(host, port)

    try:
        yield host, port
    finally:
        stop_evt.set()
        th.join(timeout=3.0)


@pytest_asyncio.fixture(scope="module")
async def ucx_server_addr():
    """
    Run UCXRemoteActorServer on a dedicated event loop in a background thread.
    Skipped entirely if UCX/CUDA/CuPy are not all available.
    """
    if not UCX_GPU_OK:
        pytest.skip("UCX/CUDA/CuPy not available")

    host = "127.0.0.1"
    with socket.socket() as s:
        s.bind((host, 0))
        port = s.getsockname()[1]

    started_evt = threading.Event()
    stop_evt = threading.Event()

    def _server_thread():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _runner():
            started_evt.set()
            await start_ucx_actor_server(host=host, port=port)

        task = loop.create_task(_runner())

        def _stop():
            if not task.done():
                task.cancel()

        def _watch_stop():
            stop_evt.wait()
            loop.call_soon_threadsafe(_stop)
            loop.call_soon_threadsafe(loop.stop)

        watcher = threading.Thread(target=_watch_stop, daemon=True)
        watcher.start()

        try:
            loop.run_forever()
        finally:
            with contextlib.suppress(asyncio.CancelledError):
                loop.run_until_complete(task)
            loop.close()

    th = threading.Thread(target=_server_thread, daemon=True)
    th.start()

    started_evt.wait(timeout=2.0)
    # actively probe UCX server
    await _wait_until_listening_ucx(host, port)

    try:
        yield host, port
    finally:
        stop_evt.set()
        th.join(timeout=3.0)


async def _make_backend(kind: str, tcp_addr: Tuple[str, int], ucx_addr: Tuple[str, int] | None):
    if kind == "thread":
        return ThreadActorBackend()
    if kind == "process":
        return ProcessActorBackend()
    if kind == "remote":
        host, port = tcp_addr
        return RemoteActorBackend(host, port)
    if kind == "gpu":
        return GPUActorBackend()
    if kind == "ucx":
        if not UCX_GPU_OK:
            pytest.skip("UCX/CUDA/CuPy not available")
        assert ucx_addr is not None
        host, port = ucx_addr
        return UCXRemoteActorBackend(host, port)
    raise RuntimeError(kind)


@pytest.mark.asyncio
async def test_gpu_basic_make_and_score_cpu_tensor():
    A = GPUActorBackend()
    await A.start()
    await A.construct(Worker, args=(), kwargs={"bias": 2.0})
    try:
        t = await A.call("make", 4)
        s = await A.call("score", t)
        assert isinstance(t, torch.Tensor) and t.shape == (4,)
        assert math.isfinite(s)
    finally:
        await A.close()


@pytest.mark.asyncio
async def test_gpu_local_channel_roundtrip_cpu_tensor():
    A = GPUActorBackend()
    await A.start()
    await A.construct(Worker, args=(), kwargs={})
    try:
        epA = await A.get_endpoint()
        chA = await open_channel(A, "gpu_local")
        payload = await A.call("make", 5)
        await chA.send(epA, payload)
        got = await chA.recv(timeout=1.0)
        assert torch.equal(got, torch.arange(5, dtype=torch.float32))
    finally:
        await A.close()


_GPU_MATRIX = [
    ("gpu", "thread"),
    ("thread", "gpu"),
    ("gpu", "process"),
    ("process", "gpu"),
    ("gpu", "remote"),
    ("remote", "gpu"),
    ("gpu", "gpu"),
]

@pytest.mark.asyncio
@pytest.mark.parametrize("kindA,kindB", _GPU_MATRIX)
async def test_gpu_cross_backend_matrix_cpu(kindA, kindB, tcp_server_addr, ucx_server_addr=None):
    A = await _make_backend(kindA, tcp_server_addr, ucx_server_addr)
    B = await _make_backend(kindB, tcp_server_addr, ucx_server_addr)

    await A.start(); await B.start()
    await A.construct(Worker, args=(), kwargs={})
    await B.construct(Worker, args=(), kwargs={})
    try:
        ch_name = f"mb_gpu_cpu_{kindA}_{kindB}_{id(A)}_{id(B)}"
        chA = await open_channel(A, ch_name)
        chB = await open_channel(B, ch_name)

        epB = await B.get_endpoint()
        payload = await A.call("make", 7)
        await chA.send(epB, payload)
        got = await chB.recv(timeout=1.0)

        assert isinstance(got, torch.Tensor)
        assert torch.equal(got, torch.arange(7, dtype=torch.float32))
    finally:
        await A.close(); await B.close()


_UCX_MATRIX = [
    ("gpu", "ucx"),
    ("ucx", "gpu"),
    ("ucx", "ucx"),
]

@pytest.mark.asyncio
@pytest.mark.parametrize("kindA,kindB", _UCX_MATRIX)
@pytest.mark.skipif(not UCX_GPU_OK, reason="UCX/CUDA/CuPy not available")
async def test_gpu_ucx_cross_backend_cuda(kindA, kindB, tcp_server_addr, ucx_server_addr):
    """
    UCX path with CUDA tensors: exercises GPU→UCX and UCX→GPU, plus UCX→UCX.
    """
    A = await _make_backend(kindA, tcp_server_addr, ucx_server_addr)
    B = await _make_backend(kindB, tcp_server_addr, ucx_server_addr)

    await A.start(); await B.start()
    await A.construct(Worker, args=(), kwargs={})
    await B.construct(Worker, args=(), kwargs={})

    try:
        ch_name = f"mb_ucx_cuda_{kindA}_{kindB}_{id(A)}_{id(B)}"
        chA = await open_channel(A, ch_name)
        chB = await open_channel(B, ch_name)

        epB = await B.get_endpoint()

        payload = torch.arange(9, dtype=torch.float32, device="cuda")
        await chA.send(epB, payload)

        got = await chB.recv(timeout=2.0)
        assert isinstance(got, torch.Tensor)
        assert torch.equal(got.to("cpu"), torch.arange(9, dtype=torch.float32))
    finally:
        await A.close(); await B.close()


@pytest.mark.asyncio
async def test_gpu_actorref_local_loopback():
    A = GPUActorBackend()
    ref = ActorRef(A)

    async with ref:
        await ref._backend.start()
        await ref._backend.construct(Worker, args=(), kwargs={"bias": 3.0})
        t = await ref.make(4)
        s = await ref.score(t)
        assert isinstance(s, float) and s > 0.0

        ch = await ref.open_channel("gpu_ref")
        ep = await ref.endpoint()
        await ch.send(ep, torch.tensor([1.0, 2.0, 3.0]))
        got = await ch.recv(timeout=1.0)
        assert torch.equal(got, torch.tensor([1.0, 2.0, 3.0]))


@pytest.mark.asyncio
@pytest.mark.skipif(not UCX_GPU_OK, reason="UCX/CUDA/CuPy not available")
async def test_gpu_actorref_ucx_roundtrip(ucx_server_addr, tcp_server_addr):
    """
    ActorRef sugar across GPU <-> UCX using a CUDA tensor over UCX.
    """
    ga = GPUActorBackend()
    ub = UCXRemoteActorBackend(*ucx_server_addr)
    refA = ActorRef(ga)
    refB = ActorRef(ub)

    async with refA, refB:
        await refA._backend.start(); await refB._backend.start()
        await refA._backend.construct(Worker, args=(), kwargs={})
        await refB._backend.construct(Worker, args=(), kwargs={})

        ch_name = f"gpu_ucx_ref_{id(ga)}_{id(ub)}"
        chA = await refA.open_channel(ch_name)
        chB = await refB.open_channel(ch_name)

        epB = await refB.endpoint()
        payload = torch.arange(11, dtype=torch.float32, device="cuda")
        await chA.send(epB, payload)

        got = await chB.recv(timeout=2.0)
        assert isinstance(got, torch.Tensor)
        assert torch.equal(got.to("cpu"), torch.arange(11, dtype=torch.float32))



from __future__ import annotations
from typing import Any, Optional, List, Iterable
import numpy as np
import torch
import torch.nn as nn

from byzpy.attacks.base import Attack
from byzpy.configs.backend import get_backend
from byzpy.engine.graph.operator import OpContext
from byzpy.engine.graph.subtask import SubTask
from byzpy.engine.storage.shared_store import (
    SharedTensorHandle,
    register_tensor,
    open_tensor,
    cleanup_tensor,
)
from byzpy.aggregators._chunking import select_adaptive_chunk_size


class SignFlipAttack(Attack):
    """g_mal = scale * base_grad (default: sign flip with scale=-1)."""

    uses_base_grad = True
    supports_subtasks = True
    max_subtasks_inflight = 0

    def __init__(self, scale: float = -1.0, *, chunk_size: int = 8192) -> None:
        self.scale = float(scale)
        if chunk_size <= 0:
            raise ValueError("chunk_size must be > 0")
        self.chunk_size = int(chunk_size)
        self._handle: SharedTensorHandle | None = None
        self._flat_shape: tuple[int, ...] | None = None
        self._like_template: Any | None = None

    def apply(
        self,
        *,
        model: Optional[nn.Module] = None,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        honest_grads: Optional[List[Any]] = None,
        base_grad: Optional[Any] = None,
    ) -> Any:
        if base_grad is None:
            raise ValueError("SignFlipAttack requires base_grad.")
        be = get_backend()
        g = be.asarray(base_grad, like=base_grad)
        out = self.scale * g
        return be.copy(out)

    def create_subtasks(self, inputs, *, context: OpContext):  # type: ignore[override]
        base_grad = inputs.get("base_grad")
        if base_grad is None:
            return []

        flat_shape, flat = _flatten_single_gradient(base_grad)
        self._flat_shape = flat_shape
        self._like_template = base_grad
        handle = register_tensor(flat)
        self._handle = handle
        total = flat.shape[0]
        metadata = getattr(context, "metadata", None) or {}
        pool_size = int(metadata.get("pool_size") or 0)
        chunk = select_adaptive_chunk_size(total, self.chunk_size, pool_size=pool_size, allow_small_chunks=True)

        def _iter() -> Iterable[SubTask]:
            chunk_id = 0
            for start in range(0, total, chunk):
                end = min(total, start + chunk)
                yield SubTask(
                    fn=_signflip_chunk,
                    args=(handle, start, end, self.scale),
                    kwargs={},
                    name=f"signflip_chunk_{chunk_id}",
                )
                chunk_id += 1

        return _iter()

    def reduce_subtasks(self, partials, inputs, *, context: OpContext):  # type: ignore[override]
        if not partials:
            return super().compute(inputs, context=context)

        if self._handle is None or self._flat_shape is None or self._like_template is None:
            raise RuntimeError("SignFlipAttack missing chunk state.")

        total = int(np.prod(self._flat_shape))
        assembled = np.zeros(total, dtype=np.float64)
        for idx, part in enumerate(partials):
            try:
                start, chunk = part
            except Exception as exc:  # pragma: no cover
                raise ValueError(f"SignFlipAttack received malformed partial at index {idx}: {part!r}") from exc
            end = start + chunk.shape[0]
            assembled[start:end] = chunk

        try:
            reshaped = assembled.reshape(self._flat_shape)
            return _to_like(reshaped, self._like_template)
        finally:
            cleanup_tensor(self._handle)
            self._handle = None
            self._flat_shape = None
            self._like_template = None


def _signflip_chunk(handle: SharedTensorHandle, start: int, end: int, scale: float) -> tuple[int, np.ndarray]:
    with open_tensor(handle) as flat:
        view = np.array(flat, copy=False)
        chunk = np.array(view[start:end], copy=True)
        chunk *= scale
        return start, chunk


def _flatten_single_gradient(grad: Any) -> tuple[tuple[int, ...], np.ndarray]:
    if isinstance(grad, SharedTensorHandle):
        with open_tensor(grad) as arr:
            np_arr = np.array(arr, copy=True)
    elif isinstance(grad, dict) and {"name", "shape", "dtype"} <= grad.keys():
        handle = SharedTensorHandle(**grad)
        with open_tensor(handle) as arr:
            np_arr = np.array(arr, copy=True)
    elif isinstance(grad, torch.Tensor):
        np_arr = grad.detach().cpu().numpy()
    else:
        np_arr = np.asarray(grad)
    shape = np_arr.shape
    flat = np_arr.reshape(-1)
    return shape, flat


def _to_like(arr: np.ndarray, like: Any) -> Any:
    if isinstance(like, torch.Tensor):
        return torch.from_numpy(arr).to(device=like.device, dtype=like.dtype)
    be = get_backend()
    return be.asarray(arr, like=like)



from __future__ import annotations
from typing import Any, Iterable, List, Optional, Sequence
import numpy as np
import torch
import torch.nn as nn

from byzpy.attacks.base import Attack
from byzpy.aggregators._chunking import select_adaptive_chunk_size
from byzpy.configs.backend import get_backend
from byzpy.engine.graph.subtask import SubTask
from byzpy.engine.storage.shared_store import (
    SharedTensorHandle,
    cleanup_tensor,
    open_tensor,
    register_tensor,
)
from byzpy.aggregators.coordinate_wise._tiling import flatten_gradients


def _to_like(arr: np.ndarray, like: Any) -> Any:
    if isinstance(like, torch.Tensor):
        return torch.from_numpy(arr).to(dtype=like.dtype, device=like.device)
    be = get_backend()
    return be.asarray(arr, like=like)


def _copy_chunk(src: SharedTensorHandle, dst: SharedTensorHandle, start: int, end: int):
    with open_tensor(src) as s, open_tensor(dst) as d:
        d[start:end] = s[start:end]
    return start, None


class MimicAttack(Attack):
    """
    Mimic an honest worker: return the vector of worker epsilon.
    """

    uses_honest_grads = True
    supports_subtasks = True
    max_subtasks_inflight = 0

    def __init__(self, epsilon: int = 0, *, chunk_size: int = 8192) -> None:
        if epsilon < 0:
            raise ValueError("epsilon must be >= 0")
        if chunk_size <= 0:
            raise ValueError("chunk_size must be > 0")
        self.epsilon = int(epsilon)
        self.chunk_size = int(chunk_size)
        self._src_handle: SharedTensorHandle | None = None
        self._dst_handle: SharedTensorHandle | None = None
        self._flat_shape: tuple[int, ...] | None = None
        self._like_template: Any | None = None

    def apply(
        self,
        *,
        model: Optional[nn.Module] = None,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        honest_grads: Optional[List[Any]] = None,
        base_grad: Optional[Any] = None,
    ) -> Any:
        if not honest_grads:
            raise ValueError("MimicAttack requires honest_grads.")
        if self.epsilon >= len(honest_grads):
            raise ValueError(f"epsilon must be < number of honest_grads (got epsilon={self.epsilon}, n={len(honest_grads)})")
        g = honest_grads[self.epsilon]
        if isinstance(g, (np.ndarray, torch.Tensor)):
            return g
        be = get_backend()
        return be.asarray(g, like=g)

    def create_subtasks(self, inputs, *, context):  # type: ignore[override]
        grads = inputs.get("honest_grads")
        if not isinstance(grads, Sequence) or not grads:
            return []
        if self.epsilon >= len(grads):
            raise ValueError(f"epsilon must be < number of honest_grads (got epsilon={self.epsilon}, n={len(grads)})")

        flat_shape, flat = flatten_gradients(grads)
        feature_dim = flat.shape[1]
        target = np.array(flat[self.epsilon], copy=True)
        src_handle = register_tensor(target)
        dst_handle = register_tensor(np.zeros_like(target))

        self._src_handle = src_handle
        self._dst_handle = dst_handle
        self._flat_shape = flat_shape
        self._like_template = grads[0]

        metadata = getattr(context, "metadata", None) or {}
        pool_size = int(metadata.get("pool_size") or 0)
        chunk = select_adaptive_chunk_size(feature_dim, self.chunk_size, pool_size=pool_size, allow_small_chunks=True)

        def _iter() -> Iterable[SubTask]:
            chunk_id = 0
            for start in range(0, feature_dim, chunk):
                end = min(feature_dim, start + chunk)
                yield SubTask(
                    fn=_copy_chunk,
                    args=(src_handle, dst_handle, start, end),
                    kwargs={},
                    name=f"mimic_chunk_{chunk_id}",
                )
                chunk_id += 1

        return _iter()

    def reduce_subtasks(self, partials, inputs, *, context):  # type: ignore[override]
        if not partials:
            return super().compute(inputs, context=context)
        if (
            self._src_handle is None
            or self._dst_handle is None
            or self._flat_shape is None
            or self._like_template is None
        ):
            raise RuntimeError("MimicAttack missing state for reduction.")

        try:
            with open_tensor(self._dst_handle) as dst:
                data = np.array(dst, copy=True)
            reshaped = data.reshape(self._flat_shape)
            return _to_like(reshaped, self._like_template)
        finally:
            cleanup_tensor(self._src_handle)
            cleanup_tensor(self._dst_handle)
            self._src_handle = None
            self._dst_handle = None
            self._flat_shape = None
            self._like_template = None


__all__ = ["MimicAttack"]



from __future__ import annotations
import asyncio
import uuid
import traceback
from typing import Any, Dict, Optional, Tuple, List
from multiprocessing import get_context
import threading

import cloudpickle

from ..base import ActorBackend
from ..channels import Endpoint
from ..router import channel_router
from ..transports import ucx, tcp
from ..ipc import wrap_payload, unwrap_payload


def _worker(conn):
    obj = None
    import queue as qmod
    mailboxes: Dict[str, qmod.Queue] = {}

    def _ensure_q(name: str) -> qmod.Queue:
        q = mailboxes.get(name)
        if q is None:
            q = qmod.Queue()
            mailboxes[name] = q
        return q

    try:
        while True:
            msg = conn.recv()
            if msg is None:
                break
            try:
                op = msg["op"]

                if op == "construct":
                    target = cloudpickle.loads(msg["blob"])
                    args  = unwrap_payload(msg["args"])
                    kwargs = unwrap_payload(msg["kwargs"])
                    obj = target(*args, **kwargs) if not isinstance(target, type) else target(*args, **kwargs)
                    conn.send({"ok": True, "id": msg["id"], "payload": None, "actor_id": msg["actor_id"]})

                elif op == "call":
                    if obj is None:
                        raise RuntimeError("ProcessActor not initialized; call construct() first.")
                    fn = getattr(obj, msg["method"])
                    args  = unwrap_payload(msg["args"])
                    kwargs = unwrap_payload(msg["kwargs"])
                    res = fn(*args, **kwargs)
                    conn.send({"ok": True, "id": msg["id"], "payload": wrap_payload(res)})

                elif op == "chan_open":
                    _ensure_q(msg["name"])
                    conn.send({"ok": True, "id": msg["id"]})

                elif op == "chan_deliver":
                    name = msg["name"]
                    payload = msg["payload"]
                    _ensure_q(name).put(payload)
                    conn.send({"ok": True, "id": msg["id"]})

                elif op == "chan_get":
                    name = msg["name"]
                    timeout = msg.get("timeout", None)
                    q = _ensure_q(name)
                    try:
                        item = q.get(timeout=timeout) if timeout is not None else q.get()
                    except qmod.Empty:
                        payload = None
                    else:
                        if isinstance(item, tuple) and len(item) == 2:
                            _, raw_payload = item
                            payload = raw_payload
                        else:
                            payload = item
                    conn.send({"ok": True, "id": msg["id"], "payload": payload})

                else:
                    raise RuntimeError(f"unknown op {op!r}")

            except Exception as e:
                conn.send({
                    "ok": False,
                    "id": msg.get("id"),
                    "payload": (type(e).__name__, str(e), traceback.format_exc()),
                })
    finally:
        try: conn.close()
        except Exception: pass


class ProcessActorBackend(ActorBackend):
    def __init__(self) -> None:
        ctx = get_context("spawn")
        parent, child = ctx.Pipe(duplex=True)
        self._parent = parent
        self._proc = ctx.Process(target=_worker, args=(child,), daemon=True)
        self._proc.start()
        child.close()

        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._actor_id = str(uuid.uuid4())
        channel_router.register("process", self._actor_id, self)
        self._io_lock = threading.Lock()

    async def start(self):
        if self._loop is None:
            self._loop = asyncio.get_running_loop()

    async def construct(self, cls_or_factory: Any, *, args: tuple, kwargs: dict) -> None:
        blob = cloudpickle.dumps(cls_or_factory)
        req = {
            "op": "construct",
            "id": str(uuid.uuid4()),
            "blob": blob,
            "args": wrap_payload(args),
            "kwargs": wrap_payload(kwargs),
            "actor_id": self._actor_id
        }
        await self._send(req)

    async def call(self, method: str, *args, **kwargs):
        req = {
            "op": "call",
            "id": str(uuid.uuid4()),
            "method": method,
            "args": wrap_payload(args),
            "kwargs": wrap_payload(kwargs),
        }
        return await self._send(req)

    async def close(self):
        try:
            def _rt_close():
                with self._io_lock:
                    try: self._parent.send(None)
                    except Exception: pass
                    try: self._parent.close()
                    except Exception: pass
            loop = self._loop or asyncio.get_running_loop()
            await loop.run_in_executor(None, _rt_close)
        except Exception:
            pass
        self._proc.join(timeout=5)
        channel_router.unregister("process", self._actor_id)

    async def get_endpoint(self) -> Endpoint:
        return Endpoint(scheme="process", address="", actor_id=self._actor_id)

    async def chan_open(self, name: str) -> Endpoint:
        await self._send({"op": "chan_open", "id": str(uuid.uuid4()), "name": name})
        return await self.get_endpoint()

    async def chan_put(self, *, from_ep: Endpoint, to_ep: Endpoint, name: str, payload: Any) -> None:
        if to_ep.scheme == "process":
            safe_payload = wrap_payload(payload)
            if to_ep.actor_id == self._actor_id:
                await self._send({"op": "chan_deliver", "id": str(uuid.uuid4()), "name": name,
                                  "payload": (from_ep, safe_payload)})
                return
            peer = channel_router.resolve("process", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {to_ep.actor_id}")
            await peer._send({"op": "chan_deliver", "id": str(uuid.uuid4()), "name": name,
                              "payload": (from_ep, safe_payload)})
            return

        if to_ep.scheme == "thread":
            peer = channel_router.resolve("thread", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return

        if to_ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port_str = to_ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                })
                tag, desc = ucx.pack_payload(payload)
                await ucx.send_payload(e, tag, desc, payload)
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)

            await ucx.call(host, int(port_str), _op)
            return

        if to_ep.scheme == "tcp":
            await tcp.chan_put(
                to_ep.address,
                from_ep=from_ep,
                to_ep=to_ep,
                name=name,
                payload=payload,
            )
            return

        if to_ep.scheme == "gpu":
            peer = channel_router.resolve("gpu", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local gpu actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=payload)
            return

        raise RuntimeError(f"ProcessActorBackend cannot route to {to_ep.scheme!r}")

    async def chan_get(self, *, ep: Endpoint, name: str, timeout: Optional[float]):
        if ep.scheme == "process" and ep.actor_id == self._actor_id:
            raw = await self._send({"op": "chan_get", "id": str(uuid.uuid4()), "name": name, "timeout": timeout})
            return unwrap_payload(raw)

        if ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port_str = ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {"op": "chan_get", "name": name, "timeout": timeout, "actor_id": ep.actor_id})
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
                if rep.get("payload_is_none", False):
                    return None
                return await ucx.recv_payload(e)

            return await ucx.call(host, int(port_str), _op)

        if ep.scheme == "tcp":
            return await tcp.chan_get(
                ep.address,
                name=name,
                timeout=timeout,
                actor_id=ep.actor_id,
            )

        raise RuntimeError("Endpoint mismatch")

    async def _send(self, msg: Dict):
        loop = (self._loop or asyncio.get_running_loop())

        def _rt():
            with self._io_lock:
                self._parent.send(msg)
                rep = self._parent.recv()
            if rep["ok"]:
                return unwrap_payload(rep.get("payload"))
            et, emsg, tb = rep["payload"]
            raise RuntimeError(f"[ProcessActor Remote {et}] {emsg}\n{tb}")

        try:
            return await loop.run_in_executor(None, _rt)
        except (EOFError, BrokenPipeError) as e:
            raise RuntimeError(
                "ProcessActor worker terminated while handling request "
                "(likely due to concurrent Pipe writes without locking; now fixed)."
            ) from e



"""
Categories 3-4: DistributedP2PHonestNode and DistributedP2PByzNode Integration Tests
"""
from __future__ import annotations

import asyncio
import pytest
import torch

from byzpy.engine.node.decentralized import DecentralizedNode
from byzpy.engine.node.distributed import DistributedHonestNode, DistributedByzantineNode
from byzpy.engine.node.context import InProcessContext
from byzpy.engine.peer_to_peer.topology import Topology
from byzpy.engine.graph.pool import ActorPoolConfig
from byzpy.aggregators.coordinate_wise import CoordinateWiseMedian
from byzpy.attacks import EmpireAttack



@pytest.mark.asyncio
async def test_distributedp2phonestnode_works_with_decentralizednode():
    """Verify DistributedP2PHonestNode can be used with DecentralizedNode."""
    from byzpy.engine.node.application import HonestNodeApplication

    class TestHonestNode(DistributedHonestNode):
        def local_honest_gradient(self, *, x, y):
            return torch.tensor([1.0, 2.0])

        def next_batch(self):
            return torch.tensor([[1.0]]), torch.tensor([0])

        def apply_server_gradient(self, grad_vec):
            pass

    node_impl = TestHonestNode(
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        aggregator=CoordinateWiseMedian(),
    )

    node = DecentralizedNode(
        node_id="node0",
        application=node_impl.application,
        context=InProcessContext(),
    )

    await node.start()

    assert node.application.has_pipeline(HonestNodeApplication.GRADIENT_PIPELINE)
    assert node.application.has_pipeline(HonestNodeApplication.AGGREGATION_PIPELINE)

    await node.shutdown()


@pytest.mark.asyncio
async def test_distributedp2phonestnode_half_step_execution():
    """Verify DistributedP2PHonestNode's half-step can be executed via DecentralizedNode."""
    from byzpy.engine.node.application import HonestNodeApplication

    class TestHonestNode(DistributedHonestNode):
        def local_honest_gradient(self, *, x, y):
            return x * 0.1 - y

        def next_batch(self):
            return torch.tensor([[1.0, 2.0]]), torch.tensor([0])

        def apply_server_gradient(self, grad_vec):
            pass

    node_impl = TestHonestNode(
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        aggregator=CoordinateWiseMedian(),
    )

    node = DecentralizedNode(
        node_id="node0",
        application=node_impl.application,
        context=InProcessContext(),
    )

    await node.start()

    x = torch.tensor([1.0, 2.0])
    y = torch.tensor([0.5, 1.0])
    result = await node.execute_pipeline(
        HonestNodeApplication.GRADIENT_PIPELINE,
        {"x": x, "y": y}
    )

    assert "honest_gradient" in result

    await node.shutdown()


@pytest.mark.asyncio
async def test_distributedp2phonestnode_aggregation_from_messages():
    """Verify DistributedP2PHonestNode aggregates gradients from neighbor messages."""
    from byzpy.engine.node.application import HonestNodeApplication

    topology = Topology.ring(3, k=1)
    received_gradients = []

    class TestHonestNode(DistributedHonestNode):
        def local_honest_gradient(self, *, x, y):
            return torch.tensor([1.0, 2.0])

        def next_batch(self):
            return torch.tensor([[1.0]]), torch.tensor([0])

        def apply_server_gradient(self, grad_vec):
            pass

    node_impl = TestHonestNode(
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        aggregator=CoordinateWiseMedian(),
    )

    from byzpy.engine.node.cluster import DecentralizedCluster
    cluster = DecentralizedCluster()

    node = await cluster.add_node(
        node_id="node0",
        application=node_impl.application,
        topology=topology,
        context=InProcessContext(),
    )

    async def gradient_handler(from_id, payload):
        received_gradients.append(payload["gradient"])

    node.register_message_handler("gradient", gradient_handler)

    cluster._update_node_id_maps()
    await cluster.start_all()
    await asyncio.sleep(0.2)

    test_gradients = [
        torch.tensor([1.0, 2.0]),  # own gradient
        torch.tensor([2.0, 3.0]),  # neighbor 1
        torch.tensor([3.0, 4.0]),  # neighbor 2
    ]

    result = await node.application.aggregate(gradients=test_gradients)

    assert isinstance(result, torch.Tensor)
    torch.testing.assert_close(result, torch.tensor([2.0, 3.0]))

    await cluster.shutdown_all()



@pytest.mark.asyncio
async def test_distributedp2pbyznode_works_with_decentralizednode():
    """Verify DistributedP2PByzNode can be used with DecentralizedNode."""
    attack = EmpireAttack()

    class TestByzNode(DistributedByzantineNode):
        def next_batch(self):
            return torch.tensor([[1.0]]), torch.tensor([0])

        def apply_server_gradient(self, grad_vec):
            pass

    node_impl = TestByzNode(
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        attack=attack,
    )

    node = DecentralizedNode(
        node_id="byz_node",
        application=node_impl.application,
        context=InProcessContext(),
    )

    await node.start()

    from byzpy.engine.node.application import ByzantineNodeApplication
    assert node.application.has_pipeline(ByzantineNodeApplication.ATTACK_PIPELINE)

    await node.shutdown()


@pytest.mark.asyncio
async def test_distributedp2pbyznode_broadcast_attack():
    """Verify DistributedP2PByzNode broadcasts malicious vectors."""
    topology = Topology.ring(2, k=1)  # Use 2 nodes so they're neighbors
    received_vectors = []

    attack = EmpireAttack()

    class TestByzNode(DistributedByzantineNode):
        def next_batch(self):
            return torch.tensor([[1.0]]), torch.tensor([0])

        def apply_server_gradient(self, grad_vec):
            pass

    node_impl = TestByzNode(
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
        attack=attack,
    )

    from byzpy.engine.node.cluster import DecentralizedCluster
    cluster = DecentralizedCluster()

    node = await cluster.add_node(
        node_id="node0",  # Use integer-like string IDs for topology
        application=node_impl.application,
        topology=topology,
        context=InProcessContext(),
    )

    from byzpy.engine.node.application import HonestNodeApplication
    honest_app = HonestNodeApplication(
        name="honest",
        actor_pool=[ActorPoolConfig(backend="thread", count=1)],
    )
    honest_node = await cluster.add_node(
        node_id="node1",  # Use integer-like string IDs for topology
        application=honest_app,
        topology=topology,
        context=InProcessContext(),
    )

    async def vector_handler(from_id, payload):
        received_vectors.append(payload["vector"])

    honest_node.register_message_handler("gradient", vector_handler)

    cluster._update_node_id_maps()
    await cluster.start_all()
    await asyncio.sleep(0.3)

    # Byzantine node broadcasts malicious vector
    template = torch.tensor([1.0, 2.0, 3.0])
    malicious = await node_impl.byzantine_gradient_async(
        honest_grads=[torch.tensor([1.0, 1.0, 1.0])]
    )

    await node.broadcast_message("gradient", {"vector": malicious})
    await asyncio.sleep(0.5)

    assert isinstance(malicious, torch.Tensor)
    assert malicious.shape == template.shape or malicious.numel() > 0
    assert len(received_vectors) >= 1

    await cluster.shutdown_all()




from __future__ import annotations
from typing import Any, Sequence, Iterable, List, Optional, Tuple
import random
import numpy as np

from .base import PreAggregator
from ..configs.backend import get_backend
from ..aggregators._chunking import select_adaptive_chunk_size
from ..engine.graph.subtask import SubTask
from ..engine.storage.shared_store import (
    SharedTensorHandle,
    register_tensor,
    open_tensor,
    cleanup_tensor,
)

try:
    import torch
    _HAS_TORCH = True
except Exception:  # pragma: no cover
    torch = None  # type: ignore
    _HAS_TORCH = False


class Bucketing(PreAggregator):
    """
    Bucketing pre-aggregator: group vectors into buckets and average.

    This pre-aggregator randomly permutes input vectors, splits them into
    consecutive buckets of a specified size, and returns the mean of each
    bucket. This reduces the number of vectors while preserving some
    statistical properties.

    Algorithm:
    1. Randomly permute the input vectors (or use provided permutation).
    2. Split into consecutive buckets of size ``bucket_size``.
    3. Return the mean of each bucket.

    Parameters
    ----------
    bucket_size : int
        Number of vectors per bucket. Must be >= 1. The last bucket may be
        smaller if the number of vectors is not divisible by bucket_size.
    feature_chunk_size : int, optional
        Size of feature chunks for parallel processing. Default is 8192.
    perm : Optional[Iterable[int]], optional
        Explicit permutation of indices. If None, a random permutation is
        generated. Must be a permutation of range(n) where n is the number
        of input vectors.
    rng : Optional[random.Random], optional
        Random number generator for shuffling. If None, a new generator is
        created.

    Examples
    --------
    >>> preagg = Bucketing(bucket_size=4)
    >>> vectors = [torch.randn(100) for _ in range(10)]
    >>> result = preagg.pre_aggregate(vectors)
    >>> len(result)  # 10 vectors -> ceil(10/4) = 3 buckets
    3
    >>> assert all(v.shape == (100,) for v in result)

    Notes
    -----
    - Output length is ceil(n / bucket_size) where n is input length.
    - Supports parallel execution via subtasks for large feature dimensions.
    - Time complexity: O(n * d) where n is number of vectors and d is
      dimension. With subtasks: O(n * d / workers).
    - Memory complexity: O(n * d) for stacking vectors.
    """
    name = "pre-agg/bucketing"
    supports_subtasks = True
    max_subtasks_inflight = 0

    def __init__(
        self,
        bucket_size: int,
        *,
        feature_chunk_size: int = 8192,
        perm: Optional[Iterable[int]] = None,
        rng: Optional[random.Random] = None,
    ) -> None:
        if bucket_size < 1:
            raise ValueError("bucket_size must be >= 1")
        self.bucket_size = int(bucket_size)
        if feature_chunk_size <= 0:
            raise ValueError("feature_chunk_size must be > 0")
        self.feature_chunk_size = int(feature_chunk_size)
        self.perm = None if perm is None else [int(i) for i in perm]
        self.rng = rng or random.Random()
        self._active_handle: SharedTensorHandle | None = None
        self._flat_shape: tuple[int, ...] | None = None
        self._bucket_slices: tuple[tuple[int, int], ...] | None = None
        self._bucket_count: int | None = None
        self._like_template: Any | None = None

    def pre_aggregate(self, xs: Sequence[Any]) -> List[Any]:
        if not xs:
            raise ValueError("xs must be a non-empty sequence")

        n = len(xs)
        be = get_backend()
        like = xs[0]

        order = self._resolve_order(n)

        arrs = [be.asarray(xs[i], like=like) for i in order]
        X = be.stack(arrs, axis=0)

        out: List[Any] = []
        for start in range(0, n, self.bucket_size):
            stop = min(start + self.bucket_size, n)
            chunk = X[start:stop]
            mean = be.mean(chunk, axis=0)
            out.append(mean)
        return out

    def create_subtasks(self, inputs, *, context):  # type: ignore[override]
        xs = inputs.get(self.input_key)
        if not isinstance(xs, Sequence) or not xs:
            return []

        n = len(xs)
        order = self._resolve_order(n)

        arrays = [_to_numpy(xs[i]) for i in order]
        stacked = np.stack(arrays, axis=0)
        self._flat_shape = stacked.shape[1:]
        flat = stacked.reshape(stacked.shape[0], -1)
        handle = register_tensor(flat)
        self._active_handle = handle
        self._like_template = xs[0]

        bucket_slices = []
        for start in range(0, n, self.bucket_size):
            stop = min(start + self.bucket_size, n)
            bucket_slices.append((start, stop))
        self._bucket_slices = tuple(bucket_slices)
        bucket_count = len(bucket_slices)
        self._bucket_count = bucket_count

        metadata = getattr(context, "metadata", None) or {}
        pool_size = int(metadata.get("pool_size") or 0)
        bucket_chunk = select_adaptive_chunk_size(
            bucket_count,
            max(1, self.feature_chunk_size),
            pool_size=pool_size,
            allow_small_chunks=True,
        )

        def _iter() -> Iterable[SubTask]:
            chunk_id = 0
            for start in range(0, bucket_count, bucket_chunk):
                end = min(bucket_count, start + bucket_chunk)
                yield SubTask(
                    fn=_bucketing_bucket_chunk,
                    args=(handle, self._bucket_slices[start:end], start),
                    kwargs={},
                    name=f"bucketing_chunk_{chunk_id}",
                )
                chunk_id += 1

        return _iter()

    def reduce_subtasks(self, partials, inputs, *, context):  # type: ignore[override]
        if not partials:
            return super().compute(inputs, context=context)

        if (
            self._active_handle is None
            or self._flat_shape is None
            or self._bucket_slices is None
            or self._bucket_count is None
            or self._like_template is None
        ):
            raise RuntimeError("Bucketing missing reduction state.")

        feature_dim = int(np.prod(self._flat_shape))
        buckets = self._bucket_count
        assembled = np.zeros((buckets, feature_dim), dtype=np.float64)

        for idx, part in enumerate(partials):
            try:
                offset, chunk = part
            except Exception as exc:  # pragma: no cover
                raise ValueError(f"Bucketing received malformed partial at index {idx}: {part!r}") from exc
            rows = chunk.shape[0]
            assembled[offset : offset + rows, :] = np.asarray(chunk, dtype=np.float64)

        results: List[Any] = []
        for bucket_idx in range(buckets):
            flat_bucket = assembled[bucket_idx]
            reshaped = flat_bucket.reshape(self._flat_shape)
            results.append(_to_like(reshaped, self._like_template))

        cleanup_tensor(self._active_handle)
        self._active_handle = None
        self._flat_shape = None
        self._bucket_slices = None
        self._bucket_count = None
        self._like_template = None

        return results

    def _resolve_order(self, n: int) -> List[int]:
        if self.perm is None:
            order = list(range(n))
            self.rng.shuffle(order)
            return order
        if len(self.perm) != n or sorted(self.perm) != list(range(n)):
            raise ValueError("perm must be a permutation of range(n)")
        return list(self.perm)


def _bucketing_bucket_chunk(
    handle: SharedTensorHandle,
    bucket_slices: Tuple[tuple[int, int], ...],
    offset: int,
) -> tuple[int, np.ndarray]:
    with open_tensor(handle) as flat:
        view = flat[:, :]
        feature_dim = view.shape[1]
        out = np.zeros((len(bucket_slices), feature_dim), dtype=view.dtype)
        for idx, (b_start, b_end) in enumerate(bucket_slices):
            chunk = view[b_start:b_end]
            out[idx, :] = np.mean(chunk, axis=0)
    return offset, out


def _to_numpy(x: Any) -> np.ndarray:
    if _HAS_TORCH and isinstance(x, torch.Tensor):  # type: ignore[arg-type]
        return x.detach().cpu().numpy()
    if isinstance(x, np.ndarray):
        return x
    return np.asarray(x)


def _to_like(arr: np.ndarray, like: Any) -> Any:
    if _HAS_TORCH and isinstance(like, torch.Tensor):  # type: ignore[arg-type]
        return torch.from_numpy(arr).to(dtype=like.dtype)
    be = get_backend()
    return be.asarray(arr, like=like)



import torch

from byzpy.attacks.inf import InfAttack
from byzpy.engine.graph.operator import OpContext


def test_inf_direct_output():
    grads = [torch.randn(3), torch.randn(3)]
    attack = InfAttack()
    out = attack.apply(honest_grads=grads)
    assert torch.isinf(out).all()


def test_inf_chunk_matches_direct():
    grads = [torch.randn(64) for _ in range(8)]
    direct = InfAttack(chunk_size=16).apply(honest_grads=grads)

    chunked = InfAttack(chunk_size=16)
    inputs = {"honest_grads": grads}
    ctx = OpContext(node_name="inf", metadata={"pool_size": 4})
    subtasks = list(chunked.create_subtasks(inputs, context=ctx))
    partials = [t.fn(*t.args, **t.kwargs) for t in subtasks]
    reduced = chunked.reduce_subtasks(partials, inputs, context=OpContext(node_name="inf", metadata={}))

    assert torch.allclose(direct, reduced)



from __future__ import annotations
from typing import Any, Dict, Optional, Protocol

from .channels import ChannelRef, Endpoint


class ActorBackend(Protocol):
    async def start(self) -> None: ...
    async def construct(self, cls_or_factory: Any, *, args: tuple, kwargs: dict) -> None: ...
    async def call(self, method: str, *args, **kwargs) -> Any: ...
    async def close(self) -> None: ...
    async def get_endpoint(self) -> Endpoint: ...
    async def chan_open(self, name: str) -> Endpoint: ...
    async def chan_put(self, *, from_ep: Endpoint, to_ep: Endpoint, name: str, payload: Any) -> None: ...
    async def chan_get(self, *, ep: Endpoint, name: str, timeout: Optional[float]) -> Any: ...


class ActorRef:
    """
    Thin async proxy: attribute access becomes an async RPC to backend.call(name, ...).
    Also exposes channel helpers.
    """
    def __init__(self, backend: ActorBackend):
        self._backend = backend

    def __getattr__(self, name: str):
        async def _remote(*args, **kwargs):
            return await self._backend.call(name, *args, **kwargs)
        return _remote

    async def __aenter__(self):
        await self._backend.start()
        return self

    async def __aexit__(self, *exc):
        await self._backend.close()
        return False

    async def open_channel(self, name: str) -> ChannelRef:
        ep = await self._backend.chan_open(name)
        return ChannelRef(self._backend, ep, name)

    async def endpoint(self) -> Endpoint:
        return await self._backend.get_endpoint()



"""
Shared registry and routing helpers for actor backends.
The existing actor backends relied on importing each other's module level
registries (e.g. ``_THREAD_REG``) to deliver messages across schemes. That
created tight coupling and import cycles.  This module centralizes the
registration of live backends and provides lightweight lookup helpers that
any backend can use without importing its peers.  It intentionally avoids
referencing concrete backend classes to remain dependency-free.
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass(frozen=True)
class BackendRecord:
    scheme: str
    actor_id: str
    backend: Any


class ChannelRouter:
    """Process-local registry for actor backends keyed by scheme/actor_id."""
    def __init__(self) -> None:
        self._backends: Dict[str, Dict[str, BackendRecord]] = {}

    def register(self, scheme: str, actor_id: str, backend: Any) -> None:
        """Record that ``backend`` owns ``(scheme, actor_id)`` locally."""
        reg = self._backends.setdefault(scheme, {})
        reg[actor_id] = BackendRecord(scheme=scheme, actor_id=actor_id, backend=backend)

    def unregister(self, scheme: str, actor_id: str) -> None:
        reg = self._backends.get(scheme)
        if not reg:
            return
        reg.pop(actor_id, None)
        if not reg:
            self._backends.pop(scheme, None)

    def resolve(self, scheme: str, actor_id: str) -> Optional[Any]:
        """Return the backend bound to ``(scheme, actor_id)`` if present."""
        reg = self._backends.get(scheme)
        if not reg:
            return None
        rec = reg.get(actor_id)
        if rec is None:
            return None
        return rec.backend

# Global singleton used by all backends
channel_router = ChannelRouter()



# byzpy/engine/actor/backends/remote.py
from __future__ import annotations
import asyncio
import cloudpickle
import traceback
import uuid
from typing import Dict, Any, Optional, Tuple

from .._wire import send_obj, recv_obj
from ..base import ActorBackend
from ..channels import Endpoint
from ..router import channel_router
from ..transports import ucx, tcp
from ..ipc import wrap_payload, unwrap_payload


class RemoteActorBackend(ActorBackend):
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = int(port)
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._actor_id: Optional[str] = None
        self._io_lock = asyncio.Lock()

    async def start(self) -> None:
        if self._reader is None or self._writer is None:
            self._reader, self._writer = await asyncio.open_connection(self._host, self._port)

    def _ensure_open(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        r, w = self._reader, self._writer
        if r is None or w is None:
            raise RuntimeError("RemoteActorBackend not started. Call await start() first.")
        return r, w

    async def close(self) -> None:
        async with self._io_lock:
            if self._writer is None or self._reader is None:
                self._actor_id = None
                return
            try:
                if self._actor_id is not None:
                    await send_obj(self._writer, {"op": "close", "actor_id": self._actor_id})
                    await recv_obj(self._reader)
            except Exception:
                pass
            finally:
                try:
                    self._writer.close()
                    await self._writer.wait_closed()
                except Exception:
                    pass
                self._reader = None
                self._writer = None
                self._actor_id = None

    async def construct(self, cls_or_factory: Any, *, args: tuple, kwargs: dict) -> None:
        async with self._io_lock:
            await self.start()
            r, w = self._ensure_open()
            blob = cloudpickle.dumps(cls_or_factory)
            await send_obj(w, {"op": "construct", "blob": blob, "args": wrap_payload(args), "kwargs": wrap_payload(kwargs)})
            rep = await recv_obj(r)
            if not rep.get("ok", False):
                et, em, tb = rep["payload"]
                raise RuntimeError(f"[RemoteActor construct {et}] {em}\n{tb}")
            self._actor_id = rep["actor_id"]

    async def call(self, method: str, *args, **kwargs):
        async with self._io_lock:
            if not self._actor_id:
                raise RuntimeError("Remote actor not constructed.")
            await self.start()
            r, w = self._ensure_open()
            await send_obj(w, {"op": "call", "actor_id": self._actor_id,
                               "method": method, "args": wrap_payload(args), "kwargs": wrap_payload(kwargs)})
            rep = await recv_obj(r)
            if rep.get("ok", False):
                return unwrap_payload(rep["payload"])
            et, em, tb = rep["payload"]
            raise RuntimeError(f"[RemoteActor call {et}] {em}\n{tb}")

    async def get_endpoint(self) -> Endpoint:
        async with self._io_lock:
            await self.start()
            r, w = self._ensure_open()
            await send_obj(w, {"op": "get_ep"})
            rep = await recv_obj(r)
            if not rep.get("ok", False):
                raise RuntimeError(rep)
            return Endpoint(**rep["payload"])

    async def chan_open(self, name: str) -> Endpoint:
        async with self._io_lock:
            await self.start()
            r, w = self._ensure_open()
            await send_obj(w, {"op": "chan_open", "name": name})
            rep = await recv_obj(r)
            if not rep.get("ok", False):
                raise RuntimeError(rep)
            return Endpoint(**rep["payload"])

    async def chan_put(self, *, from_ep: Endpoint, to_ep: Endpoint, name: str, payload: Any) -> None:
        local_payload = unwrap_payload(payload)
        if to_ep.scheme == "thread":
            peer = channel_router.resolve("thread", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=local_payload)
            return

        if to_ep.scheme == "process":
            peer = channel_router.resolve("process", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=local_payload)
            return

        if to_ep.scheme == "gpu":
            peer = channel_router.resolve("gpu", to_ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local gpu actor {to_ep.actor_id}")
            await peer.chan_put(from_ep=from_ep, to_ep=to_ep, name=name, payload=local_payload)
            return

        dest_hostport = to_ep.address
        my_hostport = f"{self._host}:{self._port}"
        wrapped = wrap_payload(payload)
        if to_ep.scheme == "tcp" and dest_hostport == my_hostport:
            async with self._io_lock:
                await self.start()
                r, w = self._ensure_open()
                await send_obj(w, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                    "payload": wrapped,
                })
                rep = await recv_obj(r)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
            return

        if to_ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port_str = to_ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {
                    "op": "chan_put",
                    "from": from_ep.__dict__,
                    "to": to_ep.__dict__,
                    "name": name,
                })
                tag, desc = ucx.pack_payload(wrapped)
                await ucx.send_payload(e, tag, desc, wrapped)
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)

            await ucx.call(host, int(port_str), _op)
            return

        if to_ep.scheme == "tcp":
            await tcp.chan_put(
                dest_hostport,
                from_ep=from_ep,
                to_ep=to_ep,
                name=name,
                payload=wrapped,
            )
            return

        raise RuntimeError(f"RemoteActorBackend cannot route to {to_ep.scheme!r}")

    async def chan_get(self, *, ep: Endpoint, name: str, timeout: Optional[float]):
        if ep.scheme == "thread":
            peer = channel_router.resolve("thread", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local thread actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)

        if ep.scheme == "process":
            peer = channel_router.resolve("process", ep.actor_id)
            if peer is None:
                raise RuntimeError(f"no local process actor {ep.actor_id}")
            return await peer.chan_get(ep=ep, name=name, timeout=timeout)

        if ep.scheme == "ucx":
            if not ucx.have_ucx():
                raise RuntimeError("UCX requested but UCX is not available (need ucxx or ucp)")
            host, port_str = ep.address.rsplit(":", 1)

            async def _op(e):
                await ucx.send_control(e, {"op": "chan_get", "name": name, "timeout": timeout, "actor_id": ep.actor_id})
                rep = await ucx.recv_control(e)
                if not rep.get("ok", False):
                    raise RuntimeError(rep)
                if rep.get("payload_is_none", False):
                    return None
                return await ucx.recv_payload(e)

            return await ucx.call(host, int(port_str), _op)

        if ep.scheme != "tcp":
            raise RuntimeError("RemoteActorBackend only resolves tcp endpoints")
        my_hostport = f"{self._host}:{self._port}"

        if ep.address == my_hostport:
            async with self._io_lock:
                await self.start()
                r, w = self._ensure_open()
            await send_obj(w, {"op": "chan_get", "name": name, "timeout": timeout})
            rep = await recv_obj(r)
            if rep.get("ok", False):
                return unwrap_payload(rep["payload"])
            raise RuntimeError(rep)

        payload = await tcp.chan_get(
            ep.address,
            name=name,
            timeout=timeout,
            actor_id=ep.actor_id,
        )
        return unwrap_payload(payload)


class RemoteActorServer:
    def __init__(self, host: str, port: int):
        self._host, self._port = host, port
        self._actors: Dict[str, Any] = {}
        self._mb: Dict[str, Dict[str, asyncio.Queue]] = {}
        self._srv: Optional[asyncio.AbstractServer] = None
        self._addr_cache: Optional[str] = None

    def _hostport(self) -> str:
        if self._addr_cache:
            return self._addr_cache
        srv = self._srv
        if srv is not None:
            sockets = getattr(srv, "sockets", None)
            if sockets:
                host, port = sockets[0].getsockname()[0], sockets[0].getsockname()[1]
                self._addr_cache = f"{host}:{port}"
                return self._addr_cache
        return f"{self._host}:{self._port}"

    async def serve(self):
        self._srv = await asyncio.start_server(self._handle, self._host, self._port)
        sockets = getattr(self._srv, "sockets", None)
        if sockets:
            host, port = sockets[0].getsockname()[0], sockets[0].getsockname()[1]
            self._addr_cache = f"{host}:{port}"
        print(f"[RemoteActorServer] Serving on {self._hostport()}")
        async with self._srv:
            await self._srv.serve_forever()

    async def _ensure_mb(self, aid: str, name: str) -> asyncio.Queue:
        chs = self._mb.setdefault(aid, {})
        q = chs.get(name)
        if q is None:
            q = asyncio.Queue()
            chs[name] = q
        return q

    async def _handle(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        bound_aid: Optional[str] = None

        try:
            while True:
                req = await recv_obj(reader)
                op = req.get("op")

                if op == "construct":
                    target = cloudpickle.loads(req["blob"])
                    args = unwrap_payload(req.get("args", ()))
                    kwargs = unwrap_payload(req.get("kwargs", {}) or {})
                    obj = target(*args, **kwargs)
                    aid = str(uuid.uuid4())
                    self._actors[aid] = obj
                    bound_aid = aid
                    await send_obj(writer, {"ok": True, "actor_id": aid})

                elif op == "get_ep":
                    if not bound_aid:
                        raise RuntimeError("no actor on this connection")
                    await send_obj(writer, {"ok": True,
                        "payload": {"scheme": "tcp", "address": self._hostport(), "actor_id": bound_aid}})

                elif op == "chan_open":
                    if not bound_aid:
                        raise RuntimeError("no actor on this connection")
                    name = req["name"]
                    await self._ensure_mb(bound_aid, name)
                    await send_obj(writer, {"ok": True,
                        "payload": {"scheme": "tcp", "address": self._hostport(), "actor_id": bound_aid}})

                elif op == "chan_put":
                    to = req["to"]; payload = req["payload"]; name = req["name"]
                    to_aid = to["actor_id"]
                    q = await self._ensure_mb(to_aid, name)
                    await q.put((req["from"], payload))
                    await send_obj(writer, {"ok": True})

                elif op == "chan_get":
                    aid = bound_aid or req.get("actor_id")
                    if not aid:
                        raise RuntimeError("actor_id missing for chan_get")
                    name = req["name"]; timeout = req.get("timeout")
                    q = await self._ensure_mb(aid, name)
                    if timeout is None:
                        fr, payload = await q.get()
                    else:
                        try:
                            fr, payload = await asyncio.wait_for(q.get(), timeout=timeout)
                        except asyncio.TimeoutError:
                            payload = None
                    await send_obj(writer, {"ok": True, "payload": payload})

                elif op == "call":
                    if not bound_aid:
                        raise RuntimeError("no actor on this connection")
                    obj = self._actors[bound_aid]
                    import inspect
                    call_args = unwrap_payload(req.get("args", ()))
                    call_kwargs = unwrap_payload(req.get("kwargs", {}) or {})
                    if inspect.iscoroutinefunction(getattr(obj, req["method"])):
                        res = await getattr(obj, req["method"])(*call_args, **call_kwargs)
                    else:
                        loop = asyncio.get_running_loop()
                        res = await loop.run_in_executor(
                            None,
                            lambda: getattr(obj, req["method"])(*call_args, **call_kwargs)
                        )
                    await send_obj(writer, {"ok": True, "payload": wrap_payload(res)})

                elif op == "close":
                    if bound_aid:
                        self._actors.pop(bound_aid, None)
                        self._mb.pop(bound_aid, None)
                    await send_obj(writer, {"ok": True})

                else:
                    await send_obj(writer, {"ok": False, "payload": ("RuntimeError", f"unknown op {op}", "")})

        except asyncio.IncompleteReadError:
            pass
        except Exception as e:
            try:
                await send_obj(writer, {"ok": False, "payload": (type(e).__name__, str(e), traceback.format_exc())})
            except Exception:
                pass
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass


async def start_actor_server(host: str, port: int):
    server = RemoteActorServer(host=host, port=port)
    await server.serve()



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing import shared_memory
from typing import Iterator, Tuple, Union

import numpy as np


@dataclass(frozen=True)
class SharedTensorHandle:
    name: str
    shape: Tuple[int, ...]
    dtype: str


SharedHandleLike = Union["SharedTensorHandle", dict]


def register_tensor(array: np.ndarray) -> SharedTensorHandle:
    arr = np.asarray(array)
    shm = shared_memory.SharedMemory(create=True, size=arr.nbytes)
    try:
        buf = np.ndarray(arr.shape, dtype=arr.dtype, buffer=shm.buf)
        buf[:] = arr
    finally:
        shm.close()
    return SharedTensorHandle(name=shm.name, shape=arr.shape, dtype=str(arr.dtype))


@contextmanager
def open_tensor(handle: SharedHandleLike) -> Iterator[np.ndarray]:
    if isinstance(handle, dict):
        handle = SharedTensorHandle(**handle)
    shm = shared_memory.SharedMemory(name=handle.name)
    try:
        arr = np.ndarray(handle.shape, dtype=np.dtype(handle.dtype), buffer=shm.buf)
        yield arr
    finally:
        shm.close()


def cleanup_tensor(handle: SharedHandleLike) -> None:
    if isinstance(handle, dict):
        handle = SharedTensorHandle(**handle)
    try:
        shm = shared_memory.SharedMemory(name=handle.name)
    except FileNotFoundError:
        return
    try:
        shm.unlink()
    finally:
        shm.close()



from __future__ import annotations

import asyncio
import pytest

from byzpy.engine.graph.graph import ComputationGraph, GraphNode, graph_input
from byzpy.engine.graph.operator import Operator
from byzpy.engine.graph.scheduler import MessageAwareNodeScheduler, NodeScheduler


class _AddOp(Operator):
    def compute(self, inputs, *, context):
        return inputs["lhs"] + inputs["rhs"]


class _NoOp(Operator):
    def compute(self, inputs, *, context):
        return None


def create_simple_graph():
    """Create a simple computation graph for testing."""
    op = _NoOp()
    node = GraphNode(name="dummy", op=op, inputs={})
    return ComputationGraph(nodes=[node], outputs=["dummy"])


def create_message_aware_scheduler(graph=None):
    """Create a MessageAwareNodeScheduler for testing."""
    if graph is None:
        graph = create_simple_graph()
    return MessageAwareNodeScheduler(graph)


# Category 1: MessageAwareNodeScheduler Core Functionality

def test_messageawarenodescheduler_can_be_created():
    """Verify MessageAwareNodeScheduler can be instantiated."""
    graph = create_simple_graph()
    scheduler = MessageAwareNodeScheduler(graph)

    assert isinstance(scheduler, NodeScheduler)
    assert scheduler.graph is graph
    assert hasattr(scheduler, 'wait_for_message')
    assert hasattr(scheduler, 'deliver_message')


@pytest.mark.asyncio
async def test_messageawarenodescheduler_wait_for_message_blocks_until_delivered():
    """Verify wait_for_message() blocks until message is delivered."""
    scheduler = create_message_aware_scheduler()

    # Start waiting task
    async def wait_task():
        return await scheduler.wait_for_message("test_msg")

    wait_fut = asyncio.create_task(wait_task())

    # Give it a moment to start waiting
    await asyncio.sleep(0.01)

    # Deliver message
    scheduler.deliver_message("test_msg", {"data": 42})

    # Wait should complete
    result = await wait_fut
    assert result == {"data": 42}


@pytest.mark.asyncio
async def test_messageawarenodescheduler_wait_for_message_uses_cache():
    """Verify wait_for_message() returns cached message immediately."""
    scheduler = create_message_aware_scheduler()

    # Deliver message first
    scheduler.deliver_message("test_msg", {"data": 100})

    # Wait should return immediately from cache
    result = await scheduler.wait_for_message("test_msg")
    assert result == {"data": 100}


@pytest.mark.asyncio
async def test_messageawarenodescheduler_wait_for_message_timeout():
    """Verify wait_for_message() raises TimeoutError on timeout."""
    scheduler = create_message_aware_scheduler()

    with pytest.raises(asyncio.TimeoutError):
        await scheduler.wait_for_message("test_msg", timeout=0.1)


@pytest.mark.asyncio
async def test_messageawarenodescheduler_deliver_message_wakes_multiple_waiters():
    """Verify deliver_message() wakes all waiting futures."""
    scheduler = create_message_aware_scheduler()

    # Create multiple waiters
    waiters = [
        asyncio.create_task(scheduler.wait_for_message("test_msg"))
        for _ in range(5)
    ]

    await asyncio.sleep(0.01)  # Let them start waiting

    # Deliver message
    scheduler.deliver_message("test_msg", {"data": 999})

    # All waiters should get the message
    results = await asyncio.gather(*waiters)
    assert all(r == {"data": 999} for r in results)


@pytest.mark.asyncio
async def test_messageawarenodescheduler_message_cache_fifo():
    """Verify message cache works in FIFO order."""
    scheduler = create_message_aware_scheduler()

    # Deliver multiple messages
    scheduler.deliver_message("test_msg", {"seq": 1})
    scheduler.deliver_message("test_msg", {"seq": 2})
    scheduler.deliver_message("test_msg", {"seq": 3})

    # Waiters should get them in order
    r1 = await scheduler.wait_for_message("test_msg")
    r2 = await scheduler.wait_for_message("test_msg")
    r3 = await scheduler.wait_for_message("test_msg")

    assert r1["seq"] == 1
    assert r2["seq"] == 2
    assert r3["seq"] == 3


@pytest.mark.asyncio
async def test_messageawarenodescheduler_multiple_message_types_independent():
    """Verify different message types are handled independently."""
    scheduler = create_message_aware_scheduler()

    # Wait for different message types
    task1 = asyncio.create_task(scheduler.wait_for_message("type1"))
    task2 = asyncio.create_task(scheduler.wait_for_message("type2"))

    await asyncio.sleep(0.01)

    # Deliver only type1
    scheduler.deliver_message("type1", {"data": 1})

    # Only task1 should complete
    result1 = await task1
    assert result1 == {"data": 1}

    # task2 should still be waiting
    assert not task2.done()

    # Deliver type2
    scheduler.deliver_message("type2", {"data": 2})
    result2 = await task2
    assert result2 == {"data": 2}


# Category 5: Complex Message-Driven Scenarios

@pytest.mark.asyncio
async def test_graph_with_multiple_message_inputs():
    """Verify graph can wait for multiple message inputs."""
    from byzpy.engine.graph.graph import GraphInput

    class _AddOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["a"] + inputs["b"]

    msg_a = GraphInput.from_message("msg_a")
    msg_b = GraphInput.from_message("msg_b")
    op = _AddOp()
    node = GraphNode(name="sum", op=op, inputs={"a": msg_a, "b": msg_b})
    graph = ComputationGraph(nodes=[node], outputs=["sum"])

    scheduler = MessageAwareNodeScheduler(graph)

    exec_task = asyncio.create_task(scheduler.run({}))
    await asyncio.sleep(0.01)

    # Deliver messages in any order
    scheduler.deliver_message("msg_b", 20)
    await asyncio.sleep(0.01)
    scheduler.deliver_message("msg_a", 10)

    result = await exec_task
    assert result["sum"] == 30


@pytest.mark.asyncio
async def test_graph_with_message_and_regular_inputs():
    """Verify graph can mix message inputs with regular inputs."""
    from byzpy.engine.graph.graph import GraphInput

    class _MultiplyOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["x"] * inputs["y"]

    msg_x = GraphInput.from_message("msg_x")
    regular_y = graph_input("y")
    op = _MultiplyOp()
    node = GraphNode(name="mult", op=op, inputs={"x": msg_x, "y": regular_y})
    graph = ComputationGraph(nodes=[node], outputs=["mult"])

    scheduler = MessageAwareNodeScheduler(graph)

    exec_task = asyncio.create_task(scheduler.run({"y": 5}))
    await asyncio.sleep(0.01)

    scheduler.deliver_message("msg_x", 4)

    result = await exec_task
    assert result["mult"] == 20


@pytest.mark.asyncio
async def test_graph_with_message_dependent_nodes():
    """Verify nodes dependent on message inputs wait correctly."""
    from byzpy.engine.graph.graph import GraphInput

    class _DoubleOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["x"] * 2

    class _AddOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["a"] + inputs["b"]

    # Node 1: waits for message
    msg_input = GraphInput.from_message("data_msg")
    double_op = _DoubleOp()
    node1 = GraphNode(name="double", op=double_op, inputs={"x": msg_input})

    # Node 2: depends on node1
    add_op = _AddOp()
    node2 = GraphNode(name="sum", op=add_op, inputs={"a": "double", "b": graph_input("bias")})

    graph = ComputationGraph(nodes=[node1, node2], outputs=["sum"])
    scheduler = MessageAwareNodeScheduler(graph)

    exec_task = asyncio.create_task(scheduler.run({"bias": 10}))
    await asyncio.sleep(0.01)

    # Deliver message
    scheduler.deliver_message("data_msg", 5)

    result = await exec_task
    assert result["sum"] == 20  # (5 * 2) + 10


@pytest.mark.asyncio
async def test_messageawarenodescheduler_concurrent_deliveries():
    """Verify concurrent message deliveries are handled correctly."""
    scheduler = create_message_aware_scheduler()

    # Create multiple waiters
    waiters = [
        asyncio.create_task(scheduler.wait_for_message(f"msg_{i}"))
        for i in range(10)
    ]

    await asyncio.sleep(0.01)

    # Deliver all messages concurrently
    for i in range(10):
        scheduler.deliver_message(f"msg_{i}", {"id": i})

    # All waiters should complete
    results = await asyncio.gather(*waiters)
    assert len(results) == 10
    assert all(r["id"] == i for i, r in enumerate(results))


@pytest.mark.asyncio
async def test_messageawarenodescheduler_wait_cancellation():
    """Verify wait_for_message() can be cancelled."""
    scheduler = create_message_aware_scheduler()

    wait_task = asyncio.create_task(scheduler.wait_for_message("test_msg"))
    await asyncio.sleep(0.01)

    # Cancel the wait
    wait_task.cancel()

    with pytest.raises(asyncio.CancelledError):
        await wait_task




import torch

from byzpy.pre_aggregators import Bucketing
from byzpy.engine.graph.operator import OpContext
from byzpy.engine.storage.shared_store import cleanup_tensor


def test_bucketing_fixed_perm():
    xs = [torch.tensor([v], dtype=torch.float32) for v in [1, 2, 3, 4, 5]]

    # permutation: [2,0,4,1,3] -> buckets size 2: [3,1], [5,2], last [4]
    perm = [2, 0, 4, 1, 3]
    agg = Bucketing(bucket_size=2, perm=perm)
    out = agg.pre_aggregate(xs)

    # expected means: [2, 3.5, 4]
    got = torch.stack(out, dim=0).reshape(-1)
    assert torch.allclose(got, torch.tensor([2.0, 3.5, 4.0]))


def test_bucketing_chunk_matches_direct():
    xs = [torch.randn(1024) for _ in range(96)]
    perm = list(range(len(xs)))

    agg_direct = Bucketing(bucket_size=8, feature_chunk_size=512, perm=perm)
    direct = agg_direct.pre_aggregate(xs)

    agg_chunk = Bucketing(bucket_size=8, feature_chunk_size=512, perm=perm)
    inputs = {"vectors": xs}
    subtasks = list(agg_chunk.create_subtasks(inputs, context=OpContext(node_name="bkt", metadata={"pool_size": 4})))
    partials = [task.fn(*task.args, **task.kwargs) for task in subtasks]
    reduced = agg_chunk.reduce_subtasks(partials, inputs, context=OpContext(node_name="bkt", metadata={}))

    def _stack(vals):
        return torch.stack([v if isinstance(v, torch.Tensor) else torch.as_tensor(v) for v in vals])

    assert torch.allclose(_stack(direct), _stack(reduced), atol=1e-6)


def test_bucketing_subtasks_scale_with_pool():
    xs = [torch.randn(65536) for _ in range(128)]
    inputs = {"vectors": xs}

    def _count(pool_size: int) -> int:
        agg = Bucketing(bucket_size=16, feature_chunk_size=4096)
        ctx = OpContext(node_name="bkt", metadata={"pool_size": pool_size})
        subtasks = list(agg.create_subtasks(inputs, context=ctx))
        handle = agg._active_handle  # type: ignore[attr-defined]
        if handle is not None:
            cleanup_tensor(handle)
            agg._active_handle = None  # type: ignore[attr-defined]
        agg._flat_shape = None  # type: ignore[attr-defined]
        agg._bucket_slices = None  # type: ignore[attr-defined]
        agg._bucket_count = None  # type: ignore[attr-defined]
        agg._like_template = None  # type: ignore[attr-defined]
        return len(subtasks)

    assert _count(8) > _count(1)



from __future__ import annotations
from typing import Any, Iterable, Sequence
import numpy as np

from ..base import Aggregator
from .._chunking import select_adaptive_chunk_size
from ...configs.backend import get_backend
from ...engine.graph.subtask import SubTask
from ...engine.storage.shared_store import (
    SharedTensorHandle,
    register_tensor,
    open_tensor,
    cleanup_tensor,
)
from ._tiling import flatten_gradients

try:  # optional torch dependency
    import torch
    _HAS_TORCH = True
except Exception:  # pragma: no cover
    torch = None  # type: ignore
    _HAS_TORCH = False


class CoordinateWiseMedian(Aggregator):
    """
    Coordinate-wise median aggregator.

    This aggregator computes the median independently for each coordinate
    (dimension) across all input gradients. It is robust to up to 50% Byzantine
    nodes when the honest gradients are well-separated.

    The median is computed coordinate-wise, meaning for each dimension d, the
    output is the median of all gradients' d-th coordinate. This makes it
    computationally efficient and suitable for high-dimensional gradients.

    The aggregator supports parallel execution via subtasks, chunking the
    computation across multiple workers for improved performance on large
    gradients.

    Parameters
    ----------
    chunk_size : int, optional
        Size of chunks for parallel processing. Larger values reduce overhead
        but may limit parallelism. Default is 8192.

    Examples
    --------
    >>> aggregator = CoordinateWiseMedian(chunk_size=4096)
    >>> gradients = [torch.randn(1000) for _ in range(10)]
    >>> result = aggregator.aggregate(gradients)
    >>> assert result.shape == (1000,)

    Notes
    -----
    - Robust to up to floor((n-1)/2) Byzantine nodes where n is the total
      number of gradients.
    - Time complexity: O(n * d) where n is number of gradients and d is
      dimension. With subtasks: O(n * d / workers).
    - Memory complexity: O(n * d) for stacking gradients.
    """
    name = "coordinate-wise-median"
    supports_subtasks = True
    max_subtasks_inflight = 0

    def __init__(self, *, chunk_size: int = 8192) -> None:
        if chunk_size <= 0:
            raise ValueError("chunk_size must be > 0")
        self.chunk_size = int(chunk_size)
        self._active_handle: SharedTensorHandle | None = None
        self._flat_shape: tuple[int, ...] | None = None

    def aggregate(self, gradients: Sequence[Any]) -> Any:
        """
        Compute coordinate-wise median of gradients.

        Parameters
        ----------
        gradients : Sequence[Any]
            Non-empty sequence of gradient tensors/arrays. All must have the
            same shape and be from the same backend (NumPy, PyTorch, etc.).

        Returns
        -------
        Any
            Aggregated gradient tensor with the same shape and backend as
            inputs. Each coordinate is the median of that coordinate across
            all input gradients.

        Raises
        ------
        ValueError
            If gradients sequence is empty.
        """
        if not gradients:
            raise ValueError("gradients must be a non-empty sequence")

        be = get_backend()
        like = gradients[0]
        arrs = [be.asarray(g, like=like) for g in gradients]
        stacked = be.stack(arrs, axis=0)  # (n, ...)
        return be.median(stacked, axis=0)

    def create_subtasks(self, inputs, *, context):  # type: ignore[override]
        gradients = inputs.get(self.input_key)
        if not isinstance(gradients, Sequence) or not gradients:
            return []

        flat_shape, flat = flatten_gradients(gradients)
        self._flat_shape = flat_shape
        handle = register_tensor(flat)
        self._active_handle = handle
        features = flat.shape[1]
        metadata = getattr(context, "metadata", None) or {}
        pool_size = int(metadata.get("pool_size") or 0)
        chunk = select_adaptive_chunk_size(features, self.chunk_size, pool_size=pool_size)

        def _iter_subtasks() -> Iterable[SubTask]:
            chunk_id = 0
            for start in range(0, features, chunk):
                end = min(features, start + chunk)
                yield SubTask(
                    fn=_median_chunk,
                    args=(handle, start, end),
                    kwargs={},
                    name=f"median_chunk_{chunk_id}",
                )
                chunk_id += 1

        return _iter_subtasks()

    def reduce_subtasks(self, partials, inputs, *, context):  # type: ignore[override]
        if not partials:
            return super().compute(inputs, context=context)

        like = inputs[self.input_key][0]
        if self._flat_shape is None:
            raise RuntimeError("CoordinateWiseMedian reduce_subtasks missing shape state.")
        feature_dim = int(np.prod(self._flat_shape))

        try:
            assembled = np.zeros(feature_dim, dtype=np.float64)
            for start, chunk in sorted(partials, key=lambda x: x[0]):
                end = start + chunk.shape[0]
                assembled[start:end] = chunk
            reshaped = assembled.reshape(self._flat_shape)
            return _to_like(reshaped, like)
        finally:
            handle = self._active_handle
            if handle is not None:
                cleanup_tensor(handle)
            self._active_handle = None
            self._flat_shape = None


def _median_chunk(handle: SharedTensorHandle, start: int, end: int) -> tuple[int, np.ndarray]:
    with open_tensor(handle) as flat:
        view = np.array(flat, copy=False)
        chunk = view[:, start:end]
        chunk_copy = np.array(chunk, copy=True)
        n = chunk_copy.shape[0]
        mid = n // 2
        if n % 2 == 1:
            med = np.partition(chunk_copy, mid, axis=0)[mid, :]
        else:
            med = np.partition(chunk_copy, mid - 1, axis=0)[mid - 1, :]
    return start, med


def _to_like(arr: np.ndarray, like: Any) -> Any:
    if _HAS_TORCH and isinstance(like, torch.Tensor):  # type: ignore[arg-type]
        return torch.from_numpy(arr).to(dtype=like.dtype)
    be = get_backend()
    return be.asarray(arr, like=like)



import asyncio
import contextlib
import math
import socket
import threading
import pytest
import pytest_asyncio
import torch

from byzpy.engine.actor.base import ActorRef
from byzpy.engine.actor.channels import open_channel
from byzpy.engine.actor.backends.thread import ThreadActorBackend
from byzpy.engine.actor.backends.process import ProcessActorBackend
from byzpy.engine.actor.backends.remote import (
    RemoteActorBackend,
    start_actor_server,
)


class Worker:
    def __init__(self, bias: float = 0.0):
        self.bias = float(bias)

    def make(self, n: int) -> torch.Tensor:
        return torch.arange(n, dtype=torch.float32).contiguous()

    def add(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return x.contiguous() + y.contiguous()

    def score(self, x: torch.Tensor) -> float:
        return float((x * x).sum().item() + self.bias)


async def _wait_until_listening(host: str, port: int, timeout: float = 3.0) -> None:
    """Poll-connect until the server socket accepts, or timeout."""
    deadline = asyncio.get_event_loop().time() + timeout
    while True:
        try:
            r, w = await asyncio.open_connection(host, port)
            w.close()
            with contextlib.suppress(Exception):
                await w.wait_closed()
            return
        except Exception as e:
            if asyncio.get_event_loop().time() >= deadline:
                raise TimeoutError(f"Server {host}:{port} did not start: {e!r}")
            await asyncio.sleep(0.05)


@pytest_asyncio.fixture(scope="module")
async def tcp_server_addr():
    """
    Run the RemoteActorServer on a dedicated event loop in a background thread.
    This avoids any scheduling/race issues with the test loop.
    """
    host = "127.0.0.1"

    with socket.socket() as s:
        s.bind((host, 0))
        port = s.getsockname()[1]

    started_evt = threading.Event()
    stop_evt = threading.Event()

    def _server_thread():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _runner():
            started_evt.set()
            try:
                await start_actor_server(host, port)
            finally:
                pass

        task = loop.create_task(_runner())

        def _stop():
            if not task.done():
                task.cancel()

        def _watch_stop():
            stop_evt.wait()
            loop.call_soon_threadsafe(_stop)
            loop.call_soon_threadsafe(loop.stop)

        watcher = threading.Thread(target=_watch_stop, daemon=True)
        watcher.start()

        try:
            loop.run_forever()
        finally:
            with contextlib.suppress(asyncio.CancelledError):
                loop.run_until_complete(task)
            loop.close()

    th = threading.Thread(target=_server_thread, daemon=True)
    th.start()

    started_evt.wait(timeout=2.0)
    await _wait_until_listening(host, port)

    try:
        yield host, port
    finally:
        stop_evt.set()
        th.join(timeout=3.0)


@pytest_asyncio.fixture(params=["thread", "process", "remote"])
async def backend_pair(request, tcp_server_addr):
    kind = request.param
    if kind == "thread":
        A = ThreadActorBackend()
        B = ThreadActorBackend()
    elif kind == "process":
        A = ProcessActorBackend()
        B = ProcessActorBackend()
    elif kind == "remote":
        host, port = tcp_server_addr
        A = RemoteActorBackend(host, port)
        B = RemoteActorBackend(host, port)
    else:
        raise RuntimeError(kind)

    await A.start()
    await B.start()
    await A.construct(Worker, args=(), kwargs={"bias": 3.0})
    await B.construct(Worker, args=(), kwargs={"bias": 5.0})

    try:
        yield A, B
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_basic_rpc_make_and_score(backend_pair):
    A, B = backend_pair
    tA = await A.call("make", 8)
    tB = await B.call("make", 8)
    assert isinstance(tA, torch.Tensor) and tA.shape == (8,)
    assert isinstance(tB, torch.Tensor) and tB.shape == (8,)
    sA = await A.call("score", tA)
    sB = await B.call("score", tB)
    ss = sum(i * i for i in range(8))
    assert math.isclose(sA, ss + 3.0, rel_tol=1e-6)
    assert math.isclose(sB, ss + 5.0, rel_tol=1e-6)


@pytest.mark.asyncio
async def test_channel_roundtrip_same_backend(backend_pair):
    A, B = backend_pair
    epB = await B.get_endpoint()
    chA = await open_channel(A, "data")
    chB = await open_channel(B, "data")

    payload = await A.call("make", 4)
    await chA.send(epB, payload)

    got = await chB.recv(timeout=1.0)
    assert isinstance(got, torch.Tensor)
    assert torch.equal(got, torch.tensor([0.0, 1.0, 2.0, 3.0]))


@pytest.mark.asyncio
async def test_channel_timeout_returns_none(backend_pair):
    _, B = backend_pair
    chB = await open_channel(B, "empty")
    out = await chB.recv(timeout=0.05)  # nobody sends
    assert out is None


@pytest.mark.asyncio
async def test_cross_backend_thread_to_process():
    A = ThreadActorBackend()
    B = ProcessActorBackend()
    await A.start(); await B.start()
    await A.construct(Worker, args=(), kwargs={})
    await B.construct(Worker, args=(), kwargs={})
    try:
        epB = await B.get_endpoint()
        chA = await open_channel(A, "x")
        chB = await open_channel(B, "x")
        payload = await A.call("make", 5)
        await chA.send(epB, payload)
        got = await chB.recv(timeout=1.0)
        assert torch.equal(got, torch.arange(5, dtype=torch.float32))
    finally:
        await A.close(); await B.close()


@pytest.mark.asyncio
async def test_cross_backend_process_to_thread():
    A = ProcessActorBackend()
    B = ThreadActorBackend()
    await A.start(); await B.start()
    await A.construct(Worker, args=(), kwargs={})
    await B.construct(Worker, args=(), kwargs={})
    try:
        epB = await B.get_endpoint()
        chA = await open_channel(A, "y")
        chB = await open_channel(B, "y")
        t = await A.call("make", 6)
        await chA.send(epB, t)
        got = await chB.recv(timeout=1.0)
        assert torch.equal(got, torch.arange(6, dtype=torch.float32))
    finally:
        await A.close(); await B.close()


@pytest.mark.asyncio
async def test_remote_backend_via_tcp_server(tcp_server_addr):
    host, port = tcp_server_addr
    A = RemoteActorBackend(host, port)
    B = RemoteActorBackend(host, port)
    await A.start()
    await B.start()
    await A.construct(Worker, args=(), kwargs={"bias": 1.5})
    await B.construct(Worker, args=(), kwargs={"bias": 2.5})

    try:
        t = await A.call("make", 3)
        s = await B.call("score", t)
        assert isinstance(s, float)

        ch_name = f"tcp_{id(A)}_{id(B)}"
        chA = await open_channel(A, ch_name)
        chB = await open_channel(B, ch_name)
        epB = await B.get_endpoint()

        await chA.send(epB, torch.tensor([10.0, 20.0]))
        got = await chB.recv(timeout=1.0)
        assert torch.equal(got, torch.tensor([10.0, 20.0]))
    finally:
        await A.close(); await B.close()


@pytest.mark.asyncio
async def test_actor_ref_context_and_proxy(tcp_server_addr):
    host, port = tcp_server_addr
    ref = ActorRef(RemoteActorBackend(host, port))

    async with ref:
        await ref._backend.start()
        await ref._backend.construct(Worker, args=(), kwargs={"bias": 7.0})
        t = await ref.make(4)
        val = await ref.score(t)
        assert isinstance(val, float) and val > 0.0

        ch = await ref.open_channel("z_ref")
        ep = await ref.endpoint()
        await ch.send(ep, torch.tensor([1.0]))
        got = await ch.recv(timeout=1.0)
        assert torch.equal(got, torch.tensor([1.0]))


async def _make_backend(kind: str, tcp_server_addr):
    if kind == "thread":
        return ThreadActorBackend()
    if kind == "process":
        return ProcessActorBackend()
    if kind == "remote":
        host, port = tcp_server_addr
        return RemoteActorBackend(host, port)
    raise RuntimeError(kind)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "kindA,kindB",
    [
        ("thread", "thread"),
        ("process", "process"),
        ("remote", "remote"),
        ("thread", "process"), ("process", "thread"),
        ("thread", "remote"),  ("remote", "thread"),
        ("process", "remote"), ("remote", "process"),
    ],
)
async def test_cross_backend_matrix(kindA, kindB, tcp_server_addr):
    A = await _make_backend(kindA, tcp_server_addr)
    B = await _make_backend(kindB, tcp_server_addr)

    await A.start(); await B.start()

    await A.construct(Worker, args=(), kwargs={})
    await B.construct(Worker, args=(), kwargs={})

    try:
        ch_name = f"mb_{kindA}_{kindB}_{id(A)}_{id(B)}"
        chA = await open_channel(A, ch_name)
        chB = await open_channel(B, ch_name)

        epB = await B.get_endpoint()

        payload = await A.call("make", 7)
        await chA.send(epB, payload)

        got = await chB.recv(timeout=1.0)
        assert isinstance(got, torch.Tensor)
        assert torch.equal(got, torch.arange(7, dtype=torch.float32))
    finally:
        await A.close(); await B.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "kindA,kindB",
    [
        ("thread", "thread"),
        ("process", "process"),
        ("remote", "remote"),
        ("thread", "process"), ("process", "thread"),
        ("thread", "remote"),  ("remote", "thread"),
        ("process", "remote"), ("remote", "process"),
    ],
)
async def test_cross_backend_matrix_actorref(kindA, kindB, tcp_server_addr):
    A = await _make_backend(kindA, tcp_server_addr)
    B = await _make_backend(kindB, tcp_server_addr)

    refA = ActorRef(A)
    refB = ActorRef(B)

    async with refA, refB:
        await refA._backend.start()
        await refB._backend.start()
        await refA._backend.construct(Worker, args=(), kwargs={})
        await refB._backend.construct(Worker, args=(), kwargs={})

        ch_name = f"mb_ref_{kindA}_{kindB}_{id(A)}_{id(B)}"
        chA = await refA.open_channel(ch_name)
        chB = await refB.open_channel(ch_name)

        epB = await refB.endpoint()

        payload = await refA.make(7)
        await chA.send(epB, payload)

        got = await chB.recv(timeout=1.0)
        assert isinstance(got, torch.Tensor)
        assert torch.equal(got, torch.arange(7, dtype=torch.float32))



"""Tests for Message Serialization - Category 4 from Milestone 5 Test Plan."""
import pytest
import torch
import numpy as np
from byzpy.engine.node.remote_client import serialize_message, deserialize_message


# =============================================================================
# Category 4.1: Basic Serialization
# =============================================================================

def test_serialize_simple_message():
    """Verify simple messages can be serialized for network transport."""
    msg = {
        "from": "node1",
        "type": "test",
        "payload": {"value": 42, "name": "test"}
    }

    serialized = serialize_message(msg)
    assert isinstance(serialized, bytes)

    deserialized = deserialize_message(serialized)
    assert deserialized == msg


def test_serialize_complex_message():
    """Verify complex nested messages can be serialized."""
    msg = {
        "from": "node1",
        "type": "complex",
        "payload": {
            "nested": {
                "list": [1, 2, 3],
                "dict": {"a": 1, "b": 2}
            }
        }
    }

    serialized = serialize_message(msg)
    deserialized = deserialize_message(serialized)

    assert deserialized == msg


# =============================================================================
# Category 4.2: Tensor Serialization
# =============================================================================

def test_serialize_tensor_message():
    """Verify tensor messages can be serialized efficiently."""
    tensor = torch.randn(100, 100)
    msg = {
        "from": "node1",
        "type": "tensor",
        "payload": {"tensor": tensor}
    }

    serialized = serialize_message(msg)
    deserialized = deserialize_message(serialized)

    assert "tensor" in deserialized["payload"]
    torch.testing.assert_close(deserialized["payload"]["tensor"], tensor)


def test_serialize_numpy_array():
    """Verify numpy arrays can be serialized."""
    array = np.random.rand(50, 50).astype(np.float32)
    msg = {
        "from": "node1",
        "type": "array",
        "payload": {"array": array}
    }

    serialized = serialize_message(msg)
    deserialized = deserialize_message(serialized)

    assert "array" in deserialized["payload"]
    np.testing.assert_array_equal(deserialized["payload"]["array"], array)


def test_serialize_large_payload():
    """Verify large payloads can be serialized efficiently."""
    large_array = np.random.rand(1000, 1000).astype(np.float32)
    msg = {
        "from": "node1",
        "type": "large",
        "payload": {"data": large_array}
    }

    serialized = serialize_message(msg)
    deserialized = deserialize_message(serialized)

    np.testing.assert_array_equal(deserialized["payload"]["data"], large_array)





from __future__ import annotations
from typing import Any, Iterable, List, Optional, Sequence
import numpy as np
import torch
import torch.nn as nn

from byzpy.attacks.base import Attack
from byzpy.aggregators._chunking import select_adaptive_chunk_size
from byzpy.configs.backend import get_backend
from byzpy.engine.graph.subtask import SubTask
from byzpy.engine.storage.shared_store import (
    SharedTensorHandle,
    register_tensor,
    open_tensor,
    cleanup_tensor,
)
from byzpy.aggregators.coordinate_wise._tiling import flatten_gradients


def _to_like(arr: np.ndarray, like: Any) -> Any:
    if isinstance(like, torch.Tensor):
        return torch.from_numpy(arr).to(dtype=like.dtype, device=like.device)
    be = get_backend()
    return be.asarray(arr, like=like)


def _make_rng(seed: Optional[int]) -> np.random.Generator:
    return np.random.default_rng(seed)


def _noop_chunk(handle: SharedTensorHandle, start: int, end: int) -> tuple[int, None]:
    # Random vector is pre-generated; chunk tasks are no-ops that keep the pipeline uniform.
    return start, None


class GaussianAttack(Attack):
    """
    Sample each coordinate independently from N(mu, sigma^2).
    """

    uses_honest_grads = True
    supports_subtasks = True
    max_subtasks_inflight = 0

    def __init__(self, mu: float = 0.0, sigma: float = 1.0, *, seed: int | None = None, chunk_size: int = 8192) -> None:
        self.mu = float(mu)
        if sigma < 0:
            raise ValueError("sigma must be >= 0")
        self.sigma = float(sigma)
        if chunk_size <= 0:
            raise ValueError("chunk_size must be > 0")
        self.chunk_size = int(chunk_size)
        self.seed = seed
        self._handle: SharedTensorHandle | None = None
        self._flat_shape: tuple[int, ...] | None = None
        self._like_template: Any | None = None

    def apply(
        self,
        *,
        model: Optional[nn.Module] = None,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        honest_grads: Optional[List[Any]] = None,
        base_grad: Optional[Any] = None,
    ) -> Any:
        if not honest_grads:
            raise ValueError("GaussianAttack requires honest_grads.")
        flat_shape, flat = flatten_gradients(honest_grads)
        feature_dim = flat.shape[1]
        rng = _make_rng(self.seed)
        sampled = rng.normal(loc=self.mu, scale=self.sigma, size=feature_dim)
        reshaped = sampled.reshape(flat_shape)
        return _to_like(reshaped, honest_grads[0])

    def create_subtasks(self, inputs, *, context):  # type: ignore[override]
        grads = inputs.get("honest_grads")
        if not isinstance(grads, Sequence) or not grads:
            return []

        flat_shape, flat = flatten_gradients(grads)
        self._flat_shape = flat_shape
        self._like_template = grads[0]
        feature_dim = flat.shape[1]

        rng = _make_rng(self.seed)
        buffer = rng.normal(loc=self.mu, scale=self.sigma, size=feature_dim).astype(np.float64, copy=False)
        handle = register_tensor(buffer)
        self._handle = handle

        metadata = getattr(context, "metadata", None) or {}
        pool_size = int(metadata.get("pool_size") or 0)
        chunk = select_adaptive_chunk_size(feature_dim, self.chunk_size, pool_size=pool_size, allow_small_chunks=True)
        starts = list(range(0, feature_dim, chunk))

        def _iter() -> Iterable[SubTask]:
            for idx, start in enumerate(starts):
                end = min(feature_dim, start + chunk)
                yield SubTask(
                    fn=_noop_chunk,
                    args=(handle, start, end),
                    kwargs={},
                    name=f"gaussian_chunk_{idx}",
                )

        return _iter()

    def reduce_subtasks(self, partials, inputs, *, context):  # type: ignore[override]
        if not partials:
            return super().compute(inputs, context=context)
        if self._handle is None or self._flat_shape is None or self._like_template is None:
            raise RuntimeError("GaussianAttack missing state for reduction.")

        try:
            with open_tensor(self._handle) as flat:
                data = np.array(flat, copy=True)
            reshaped = data.reshape(self._flat_shape)
            return _to_like(reshaped, self._like_template)
        finally:
            cleanup_tensor(self._handle)
            self._handle = None
            self._flat_shape = None
            self._like_template = None


__all__ = ["GaussianAttack"]



from __future__ import annotations

import asyncio
import pytest

from byzpy.engine.graph.graph import ComputationGraph, GraphNode, graph_input, GraphInput
from byzpy.engine.graph.operator import Operator
from byzpy.engine.graph.scheduler import MessageAwareNodeScheduler


# Category 3: GraphInput Message Sources

def test_graphinput_from_message_creates_message_source():
    """Verify GraphInput.from_message() creates a message source."""
    msg_input = GraphInput.from_message("test_msg")
    assert hasattr(msg_input, 'message_type')
    assert msg_input.message_type == "test_msg"
    assert msg_input.field is None


def test_graphinput_from_message_with_field():
    """Verify GraphInput.from_message() can extract specific field."""
    msg_input = GraphInput.from_message("test_msg", field="data")
    assert msg_input.message_type == "test_msg"
    assert msg_input.field == "data"


@pytest.mark.asyncio
async def test_graphinput_from_message_in_graph_execution():
    """Verify graph with message input waits for message."""
    class _DoubleOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["x"] * 2

    # Create graph with message input
    msg_input = GraphInput.from_message("data_msg")
    op = _DoubleOp()
    node = GraphNode(name="double", op=op, inputs={"x": msg_input})
    graph = ComputationGraph(nodes=[node], outputs=["double"])

    scheduler = MessageAwareNodeScheduler(graph)

    # Start execution
    exec_task = asyncio.create_task(scheduler.run({}))
    await asyncio.sleep(0.01)

    # Deliver message
    scheduler.deliver_message("data_msg", 5)

    # Graph should complete
    result = await exec_task
    assert result["double"] == 10


@pytest.mark.asyncio
async def test_graphinput_from_message_extracts_field():
    """Verify GraphInput.from_message() extracts specific field from message."""
    class _IdentityOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["x"]

    msg_input = GraphInput.from_message("test_msg", field="value")
    op = _IdentityOp()
    node = GraphNode(name="identity", op=op, inputs={"x": msg_input})
    graph = ComputationGraph(nodes=[node], outputs=["identity"])

    scheduler = MessageAwareNodeScheduler(graph)

    exec_task = asyncio.create_task(scheduler.run({}))
    await asyncio.sleep(0.01)

    # Deliver message with nested structure
    scheduler.deliver_message("test_msg", {"value": 42, "other": "ignored"})

    result = await exec_task
    assert result["identity"] == 42


@pytest.mark.asyncio
async def test_graphinput_from_message_missing_field():
    """Verify GraphInput.from_message() handles missing field gracefully."""
    class _IdentityOp(Operator):
        def compute(self, inputs, *, context):
            return inputs["x"]

    msg_input = GraphInput.from_message("test_msg", field="nonexistent")
    op = _IdentityOp()
    node = GraphNode(name="identity", op=op, inputs={"x": msg_input})
    graph = ComputationGraph(nodes=[node], outputs=["identity"])

    scheduler = MessageAwareNodeScheduler(graph)

    exec_task = asyncio.create_task(scheduler.run({}))
    await asyncio.sleep(0.01)

    # Deliver message without the field
    scheduler.deliver_message("test_msg", {"other": "data"})

    # Should raise KeyError or return None
    with pytest.raises((KeyError, TypeError)):
        await exec_task

